# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# flake8: noqa: E501
# pylint: disable=too-many-try-statements,broad-exception-caught
# pylint: disable=line-too-long,duplicate-code,unused-argument
# pylint: disable=too-many-arguments,too-many-positional-arguments
# pylint: disable=too-many-locals,too-many-instance-attributes

# pyright: reportMissingTypeStubs=false,reportUnknownMemberType=false
# pyright: reportUnusedParameter=false

"""An MQTT I/O stream for handling print and input messages."""

import json
import logging
import time
import traceback as tb
import uuid
from pathlib import Path
from threading import Event, Lock
from types import TracebackType
from typing import Any, Callable

try:
    from paho.mqtt import client as mqtt
    from paho.mqtt.enums import CallbackAPIVersion
    from paho.mqtt.reasoncodes import ReasonCode
except ImportError as error:  # pragma: no cover
    raise ImportError(
        "MQTT client not installed. Please install paho-mqtt with `pip install paho-mqtt`."
    ) from error

from autogen.events import BaseEvent  # type: ignore
from autogen.io import IOStream  # type: ignore
from autogen.messages import BaseMessage  # type: ignore

from .models import (
    PrintMessage,
    TextMediaContent,
    UserInputData,
    UserInputRequest,
    UserResponse,
)
from .utils import gen_id, get_message_dump, now

LOG = logging.getLogger(__name__)


MQTT_FIRST_RECONNECT_DELAY = 1
MQTT_RECONNECT_RATE = 2
MQTT_MAX_RECONNECT_COUNT = 12
MQTT_MAX_RECONNECT_DELAY = 60


# noinspection PyUnusedLocal,PyBroadException
class MqttIOStream(IOStream):
    """MQTT I/O stream."""

    client: mqtt.Client
    task_id: str
    input_timeout: int
    on_input_request: Callable[[str, str, str], None] | None
    on_input_response: Callable[[str, str], None] | None
    max_retain_messages: int
    output_topic: str
    input_request_topic: str
    input_response_topic: str
    common_output_topic: str
    broker_host: str
    broker_port: int

    # Thread safety and input handling
    _input_responses: dict[str, str]
    _input_lock: Lock
    _input_events: dict[str, Event]
    _processed_requests: set[str]
    _connected: bool

    def __init__(
        self,
        broker_host: str = "localhost",
        broker_port: int = 1883,
        task_id: str | None = None,
        input_timeout: int = 120,
        connect_timeout: int = 10,
        max_retain_messages: int = 1000,
        on_input_request: Callable[[str, str, str], None] | None = None,
        on_input_response: Callable[[str, str], None] | None = None,
        mqtt_client_kwargs: dict[str, Any] | None = None,
        uploads_root: Path | str | None = None,
        username: str | None = None,
        password: str | None = None,
        use_tls: bool = False,
        ca_cert_path: str | None = None,
    ) -> None:
        """Initialize the MQTT I/O stream.

        Parameters
        ----------
        broker_host : str, optional
            The MQTT broker host, by default "localhost".
        broker_port : int, optional
            The MQTT broker port, by default 1883.
        task_id : str, optional
            An ID to use for the topics. If not provided, a random UUID will be generated.
        input_timeout : int, optional
            The time to wait for user input in seconds, by default 120.
        connect_timeout : int, optional
            The time to wait for MQTT connection in seconds, by default 10.
        on_input_request : Optional[Callable[[str, str, str], None]], optional
            Callback for input request, by default None
            parameters: prompt, request_id, task_id
        on_input_response : Optional[Callable[[str, str], None]], optional
            Callback for input response, by default None.
            parameters: user_input, task_id
        mqtt_client_kwargs : dict[str, Any] | None, optional
            Additional MQTT client kwargs, by default None.
        max_retain_messages : int, optional
            Maximum number of retained messages per topic, by default 1000.
        uploads_root : Path | str | None, optional
            The root directory for uploads, by default None.
        username : str | None, optional
            MQTT broker username, by default None.
        password : str | None, optional
            MQTT broker password, by default None.
        use_tls : bool, optional
            Whether to use TLS connection, by default False.
        ca_cert_path : str | None, optional
            Path to CA certificate file for TLS, by default None.
        """
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.task_id = task_id or uuid.uuid4().hex
        self.input_timeout = input_timeout
        self.connect_timeout = connect_timeout
        self.on_input_request = on_input_request
        self.on_input_response = on_input_response
        self.max_retain_messages = max_retain_messages

        # Topic structure
        self.output_topic = f"task/{self.task_id}/output"
        self.input_request_topic = f"task/{self.task_id}/input_request"
        self.input_response_topic = f"task/{self.task_id}/input_response"
        self.common_output_topic = "task/output"

        # Thread safety
        self._input_responses = {}
        self._input_lock = Lock()
        self._input_events = {}
        self._processed_requests = set()
        self._connected = False

        # Uploads
        self.uploads_root = (
            Path(uploads_root).resolve() if uploads_root else None
        )
        if self.uploads_root and not self.uploads_root.exists():
            self.uploads_root.mkdir(parents=True, exist_ok=True)

        # Initialize MQTT client
        client_kwargs = mqtt_client_kwargs or {}
        if "callback_api_version" not in client_kwargs:  # pragma: no branch
            client_kwargs["callback_api_version"] = CallbackAPIVersion.VERSION2
        self.client = mqtt.Client(**client_kwargs)

        # Set up authentication
        if username and password:
            self.client.username_pw_set(username, password)

        # Set up TLS
        if use_tls:
            if ca_cert_path:
                self.client.tls_set(ca_cert_path)
            else:  # pragma: no cover
                self.client.tls_set()

        # Set up callbacks
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message
        self.client.on_log = self._on_log

        # Connect to broker
        self._connect()

    def _connect(self) -> None:
        """Connect to MQTT broker."""
        try:
            LOG.debug(
                "Connecting to MQTT broker at %s:%d",
                self.broker_host,
                self.broker_port,
            )
            self.client.connect(self.broker_host, self.broker_port, 60)
            self.client.loop_start()

            # Wait for connection
            timeout = self.connect_timeout  # seconds
            start_time = time.time()
            while (
                not self.client.is_connected()
                and (time.time() - start_time) < timeout
            ):
                time.sleep(0.1)

            if not self.client.is_connected():
                raise ConnectionError(
                    "Failed to connect to MQTT broker within timeout"
                )

        except Exception as e:
            LOG.error("Failed to connect to MQTT broker: %s", e)
            raise

    def _on_connect(
        self,
        client: mqtt.Client,
        userdata: Any,
        flags: dict[str, Any],
        reason_code: ReasonCode | int,
    ) -> None:
        """Handle MQTT connection event.

        Parameters
        ----------
        client : mqtt.Client
            The MQTT client instance.
        userdata : Any
            User-defined data of any type (not used here).
        flags : dict[str, Any]
            Response flags from the broker.
        reason_code : ReasonCode | int
            The connection reason code.
        """
        if isinstance(reason_code, ReasonCode):  # pragma: no cover
            failed = reason_code.is_failure
        else:
            failed = reason_code != mqtt.MQTT_ERR_SUCCESS
        if not failed and client.is_connected():
            LOG.debug("Connected to MQTT broker successfully")
            self._connected = True

            # Subscribe to input response topic
            client.subscribe(self.input_response_topic, qos=1)
            LOG.debug(
                "Subscribed to input response topic: %s",
                self.input_response_topic,
            )
        else:
            LOG.error(
                "Failed to connect to MQTT broker: %s (code %s)",
                userdata,
                reason_code,
            )
            self._connected = False
            raise ConnectionError(
                f"MQTT connection failed with reason code {reason_code}"
            )

    def _on_disconnect(
        self, client: mqtt.Client, userdata: Any, reason_code: ReasonCode | int
    ) -> None:
        """Handle MQTT disconnection event.

        Parameters
        ----------
        client : mqtt.Client
            The MQTT client instance.
        userdata : Any
            User-defined data of any type (not used here).
        reason_code : ReasonCode | int
            The disconnection reason code.
        """
        self._connected = False
        if isinstance(reason_code, ReasonCode):  # pragma: no cover
            is_normal_disconnect = reason_code.value == mqtt.MQTT_ERR_SUCCESS
        else:
            is_normal_disconnect = reason_code == mqtt.MQTT_ERR_SUCCESS
        if is_normal_disconnect:  # pragma: no cover
            LOG.debug("Disconnected from MQTT broker normally")
        else:
            LOG.warning("Disconnected with reason: %s", str(reason_code))
            reconnect_count, reconnect_delay = 0, MQTT_FIRST_RECONNECT_DELAY
            while reconnect_count < MQTT_MAX_RECONNECT_COUNT:
                LOG.info("Reconnecting in %d seconds...", reconnect_delay)
                time.sleep(reconnect_delay)
                # pylint: disable=broad-exception-caught
                try:
                    client.reconnect()
                except Exception as err:
                    LOG.error("%s. Reconnect failed. Retrying...", err)
                else:  # pragma: no cover
                    LOG.info("Reconnected successfully!")
                    return

                reconnect_delay *= MQTT_RECONNECT_RATE
                reconnect_delay = min(reconnect_delay, MQTT_MAX_RECONNECT_DELAY)
                reconnect_count += 1
            LOG.info("Reconnect failed after %s attempts.", reconnect_count)

    # noinspection PyUnresolvedReferences
    def _on_message(
        self, client: mqtt.Client, userdata: Any, msg: mqtt.MQTTMessage
    ) -> None:
        """Handle incoming MQTT messages.

        Parameters
        ----------
        client : mqtt.Client
            The MQTT client instance.
        userdata : Any
            User-defined data of any type (not used here).
        msg : mqtt.MQTTMessage
            The received MQTT message.
        """
        try:
            LOG.debug(
                "Received message on topic %s: %s",
                msg.topic,
                msg.payload.decode("utf-8", errors="replace"),
            )

            if msg.topic == self.input_response_topic:  # pragma: no branch
                self._handle_input_response(
                    msg.payload.decode("utf-8", errors="replace")
                )

        except Exception as e:  # pragma: no cover
            LOG.error("Error handling message: %s", e)

    def _on_log(
        self,
        client: mqtt.Client,
        userdata: Any,
        level: int,
        buf: str,
    ) -> None:  # pragma: no cover
        """Handle MQTT log messages.

        Parameters
        ----------
        client : mqtt.Client
            The MQTT client instance.
        userdata : Any
            User-defined data of any type (not used here).
        level : int
            The log level.
        buf : str
            The log message.
        """
        payload: dict[str, Any] = {
            "level": level,
            "message": buf,
        }
        LOG.debug("MQTT log: %s", payload)
        print_message = PrintMessage(data=buf)
        try:
            payload = print_message.model_dump(mode="json")
        except Exception:
            payload = print_message.model_dump(
                serialize_as_any=True, mode="json", fallback=str
            )
        self._print_to_common_output(payload=payload)

    def _handle_input_response(self, payload: str) -> None:
        """Handle input response message."""
        try:
            message_data = json.loads(payload)
            response = self._create_user_response(message_data)

            if not response or not response.request_id:
                return

            # Check if already processed
            if response.request_id in self._processed_requests:
                return

            with self._input_lock:
                self._processed_requests.add(response.request_id)
                user_input = self._get_user_input(response)
                self._input_responses[response.request_id] = user_input

                # Signal waiting thread
                if (
                    response.request_id in self._input_events
                ):  # pragma: no branch
                    self._input_events[response.request_id].set()

        except Exception as e:
            LOG.error("Error handling input response: %s", e)

    def __enter__(self) -> "MqttIOStream":
        """Enable context manager usage."""
        return self

    def __exit__(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager."""
        self.close()

    def close(self) -> None:
        """Close the MQTT client."""
        if hasattr(self, "client"):  # pragma: no branch
            try:
                self.client.loop_stop()
                self.client.disconnect()
            except Exception as e:
                LOG.error("Error closing MQTT client: %s", e)

    def _publish_message(
        self, topic: str, payload: dict[str, Any], retain: bool = False
    ) -> None:
        """Publish message to MQTT topic.

        Parameters
        ----------
        topic : str
            The MQTT topic.
        payload : dict[str, Any]
            The message payload.
        retain : bool, optional
            Whether to retain the message, by default False.
        """
        try:
            json_payload = json.dumps(payload)
            LOG.debug("Publishing to %s: %s", topic, json_payload)

            result = self.client.publish(
                topic, json_payload, qos=1, retain=retain
            )

            if result.rc != mqtt.MQTT_ERR_SUCCESS:
                LOG.error(
                    "Failed to publish message to %s: %s", topic, result.rc
                )
        except Exception as e:
            LOG.error("Error publishing message: %s", e)

    def _print_to_task_output(self, payload: dict[str, Any]) -> None:
        """Print message to the task output topic."""
        self._publish_message(self.output_topic, payload, retain=True)

    def _print_to_common_output(self, payload: dict[str, Any]) -> None:
        """Print message to the common output topic."""
        self._publish_message(self.common_output_topic, payload, retain=False)

    def _print(self, payload: dict[str, Any]) -> None:
        """Print message to MQTT topics."""
        if "id" not in payload:
            payload["id"] = gen_id()
        payload["task_id"] = self.task_id
        if "timestamp" not in payload:
            payload["timestamp"] = now()

        self._print_to_task_output(payload)
        self._print_to_common_output(payload)

    def print(self, *args: Any, **kwargs: Any) -> None:
        """Print message to MQTT topics.

        Parameters
        ----------
        args : Any
            The message to print.
        kwargs : Any
            Additional keyword arguments.
        """
        print_message = PrintMessage.create(*args, **kwargs)
        try:
            payload = print_message.model_dump(mode="json")
        except Exception:  # pragma: no cover
            payload = print_message.model_dump(
                serialize_as_any=True, mode="json", fallback=str
            )
        self._print(payload)

    def send(self, message: BaseEvent | BaseMessage) -> None:
        """Send a structured message to MQTT.

        Parameters
        ----------
        message : BaseEvent | BaseMessage
            The message or event to send.
        """
        message_dump = get_message_dump(message)
        message_type = message_dump.get("type", None)
        if not message_type:  # pragma: no cover
            message_type = message.__class__.__name__
        self._print(
            {
                "type": message_type,
                "data": json.dumps(message_dump),
            }
        )

    def input(
        self,
        prompt: str = "",
        *,
        password: bool = False,
        request_id: str | None = None,
    ) -> str:
        """Request input via MQTT and wait for response.

        Parameters
        ----------
        prompt : str, optional
            The prompt message, by default "".
        password : bool, optional
            Whether input is masked, by default False.
        request_id : str, optional
            The request ID (for testing), by default None.

        Returns
        -------
        str
            The received user input, or empty string if timeout occurs.
        """
        request_id = request_id or gen_id()

        input_request = UserInputRequest(
            request_id=request_id,
            prompt=prompt,
            password=password,
        )
        try:
            payload = input_request.model_dump(mode="json")
        except Exception:  # pragma: no cover
            payload = input_request.model_dump(
                serialize_as_any=True, mode="json", fallback=str
            )
        payload["task_id"] = self.task_id
        payload["password"] = str(password).lower()

        LOG.debug("Requesting input via MQTT: %s", payload)

        # Create event for this request
        with self._input_lock:
            self._input_events[request_id] = Event()

        # Publish input request
        self._print(payload)
        self._publish_message(self.input_request_topic, payload)

        if self.on_input_request:
            self.on_input_request(prompt, request_id, self.task_id)

        user_input = self._wait_for_input(request_id)

        if self.on_input_response:
            self.on_input_response(user_input, self.task_id)

        # Send response confirmation
        text_response = UserInputData(
            content=TextMediaContent(text=user_input),
        )
        user_response = UserResponse(
            request_id=request_id,
            type="input_response",
            data=text_response,
        )

        payload = user_response.model_dump(mode="json")
        payload["task_id"] = self.task_id
        payload["data"] = json.dumps(payload["data"])

        LOG.debug("Sending input response: %s", payload)
        self._print(payload)

        return user_input

    def _wait_for_input(self, request_id: str) -> str:
        """Wait for user input.

        Parameters
        ----------
        request_id : str
            The request ID.

        Returns
        -------
        str
            The user input.
        """
        try:
            # Wait for response
            event = self._input_events.get(request_id)
            if not event:
                LOG.error("No event found for request %s", request_id)
                return ""

            if event.wait(timeout=self.input_timeout):
                # Got response
                with self._input_lock:
                    user_input = self._input_responses.pop(request_id, "")
                    self._input_events.pop(request_id, None)
                return user_input
            # Timeout
            LOG.warning(
                "No input received for %ds on task %s, assuming empty string",
                self.input_timeout,
                self.task_id,
            )
            with self._input_lock:
                self._input_events.pop(request_id, None)
            return ""

        except Exception as e:
            LOG.error("Error in _wait_for_input: %s", e)
            return ""

    def _get_user_input(self, response: UserResponse) -> str:
        """Get user input from the response.

        Parameters
        ----------
        response : UserResponse
            The user response.

        Returns
        -------
        str
            The user input.
        """
        if not response.data:
            return ""
        if isinstance(response.data, str):  # pragma: no cover
            return response.data
        return response.to_string(
            uploads_root=self.uploads_root,
            base_name=response.request_id,
        )

    @staticmethod
    def _create_user_response(
        message_data: dict[str, Any],
    ) -> UserResponse | None:
        """Create UserResponse object from validated data."""
        try:
            # Handle nested JSON in 'data' field
            if "data" in message_data and isinstance(message_data["data"], str):
                try:
                    message_data["data"] = json.loads(message_data["data"])
                except json.JSONDecodeError:
                    LOG.error(
                        "Invalid JSON in nested data field: %s", message_data
                    )
                    return None

            return UserResponse.model_validate(message_data)
        except Exception as e:
            LOG.error(
                "Error parsing user input response: %s - %s",
                message_data,
                str(e),
            )
            return None

    @staticmethod
    def try_do(func: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Try to execute a function without raising exceptions.

        Parameters
        ----------
        func : Callable[..., Any]
            The function to call.
        args : Any
            The function's positional arguments.
        kwargs : Any
            The function's keyword arguments.
        """
        try:
            func(*args, **kwargs)
        except BaseException:  # pragma: no cover
            LOG.error("Error on try_do:")
            LOG.error(tb.format_exc())

    def cleanup_task_data(self) -> None:
        """Clean up task-specific data.

        Note: MQTT doesn't have built-in cleanup like Redis streams.
        This method clears local state and can be extended for broker-specific cleanup.
        """
        with self._input_lock:
            self._input_responses.clear()
            self._input_events.clear()
            self._processed_requests.clear()

        LOG.debug("Cleaned up task data for %s", self.task_id)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods, no-self-use
"""Processor for generating Waldiez agent definitions."""

import json
from dataclasses import dataclass
from typing import Callable

from waldiez.models import WaldiezAgent

from ..core.enums import GroupManagerStrategy
from ..core.extras.agent_extras import GroupManagerExtras, StandardExtras


@dataclass
class AgentProcessingResult:
    """Result from processing an agent."""

    content: str = ""


class AgentProcessor:
    """Processor for main agent definition generation."""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_names: dict[str, str],
        arguments_resolver: Callable[[WaldiezAgent], list[str]],
        extras: StandardExtras,
    ):
        self.agent = agent
        self.agent_names = agent_names
        self.arguments_resolver = arguments_resolver
        self.extras = extras

    def _should_skip_group_manager(self) -> bool:
        """Check if the group manager should be skipped.

        Returns
        -------
        bool
            True if the group manager should be skipped, False otherwise.
        """
        return (
            isinstance(self.extras, GroupManagerExtras)
            and self.extras.strategy == GroupManagerStrategy.PATTERN
        )

    def process(self) -> AgentProcessingResult:
        """Process the agent and generate its definition.

        Returns
        -------
        AgentProcessingResult
            The result containing the agent definition string.
        """
        if (
            isinstance(self.extras, GroupManagerExtras)
            and self._should_skip_group_manager()
        ):
            return AgentProcessingResult(content="")
        agent_name = self.agent_names[self.agent.id]

        ag2_class = self.agent.ag2_class
        # Build extra arguments
        extra_args = ""
        for arg in self.extras.extra_args:
            extra_args += arg.get_content(append_new_line=True)

        # Get additional arguments from resolver
        resolved_args = self.arguments_resolver(self.agent)
        if resolved_args:  # pragma: no branch
            extra_args += ",\n".join(resolved_args)
        if not extra_args.startswith("\n"):
            extra_args = "\n" + extra_args
        if not extra_args.endswith("\n"):
            extra_args += "\n"
        # Build the agent definition
        agent_str = self.get_agent_string(
            agent_name=agent_name, ag2_class=ag2_class, extra_args=extra_args
        )

        return AgentProcessingResult(content=agent_str)

    def get_agent_string(
        self, agent_name: str, ag2_class: str, extra_args: str
    ) -> str:
        """Get the agent definition string.

        Parameters
        ----------
        agent_name : str
            The name of the agent.
        ag2_class : str
            The class of the agent.
        extra_args : str
            Additional arguments for the agent.

        Returns
        -------
        str
            The agent definition string.
        """
        agent_str: str = f"{agent_name} = {ag2_class}(\n"
        if "name" not in self.agent.args_to_skip:
            agent_str += self.get_name_arg()
        if "description" not in self.agent.args_to_skip:
            agent_str += self.get_description_arg()
        if "system_message" not in self.agent.args_to_skip:
            agent_str += self.get_system_message_arg()
        if "human_input_mode" not in self.agent.args_to_skip:
            agent_str += self.get_human_input_mode_arg()
        if "max_consecutive_auto_reply" not in self.agent.args_to_skip:
            agent_str += self.get_max_consecutive_auto_reply_arg()
        if "default_auto_reply" not in self.agent.args_to_skip:
            agent_str += self.get_auto_reply_arg()
        if "code_execution_config" not in self.agent.args_to_skip:
            agent_str += self.get_code_execution_arg()
        if "is_termination_msg" not in self.agent.args_to_skip:
            agent_str += self.get_termination_arg()
        if extra_args:  # pragma: no branch
            agent_str = self.add_extra_args(agent_str, extra_args)
        agent_str += ")"
        return agent_str

    def get_name_arg(self) -> str:
        """Get the agent name argument.

        Returns
        -------
        str
            The agent name argument.
        """
        return f'    name="{self.agent_names[self.agent.id]}",\n'

    def get_system_message_arg(self) -> str:
        """Get the system message argument.

        Returns
        -------
        str
            The system message argument.
        """
        arg = str(self.extras.get_system_message_arg())
        if arg:
            return f"{arg},\n"
        return ""

    def get_code_execution_arg(self) -> str:
        """Get the code execution argument.

        Returns
        -------
        str
            The code execution argument.
        """
        arg = str(self.extras.get_code_execution_arg())
        if arg:
            return f"{arg},\n"
        return ""

    def get_termination_arg(self) -> str:
        """Get the termination argument.

        Returns
        -------
        str
            The termination argument.
        """
        arg = str(self.extras.get_termination_arg())
        if arg:
            return f"{arg}\n"
        return ""

    def get_human_input_mode_arg(self) -> str:
        """Get the human input mode argument.

        Returns
        -------
        str
            The human input mode argument.
        """
        return f'    human_input_mode="{self.agent.data.human_input_mode}",\n'

    def get_max_consecutive_auto_reply_arg(self) -> str:
        """Get the maximum consecutive auto reply argument.

        Returns
        -------
        str
            The maximum consecutive auto reply argument.
        """
        value = self.agent.data.max_consecutive_auto_reply
        return f"    max_consecutive_auto_reply={value},\n"

    def get_description_arg(self) -> str:
        """Get the agent description.

        Returns
        -------
        str
            The agent description.
        """
        description = (
            json.dumps(self.agent.description)
            if self.agent.description
            else '""'
        )
        return f"    description={description},\n"

    def get_auto_reply_arg(self) -> str:
        """Get the default auto reply argument.

        Returns
        -------
        str
            The default auto reply argument.
        """
        default_auto_reply = '""'
        if self.agent.data.agent_default_auto_reply:
            # Escape the default auto reply string
            default_auto_reply = json.dumps(
                self.agent.data.agent_default_auto_reply
            )
        return f"    default_auto_reply={default_auto_reply},\n"

    # noinspection PyMethodMayBeStatic
    def add_extra_args(self, agent_str: str, extra_args: str) -> str:
        """Add extra agent args.

        Parameters
        ----------
        agent_str : str
            The current agent string
        extra_args : str
            The extra args to add

        Returns
        -------
        str
            The new agent string
        """
        agent_string = str(agent_str)
        if agent_string.endswith("\n"):
            agent_string = agent_string[: -len("\n")]
        agent_string += extra_args
        return agent_string



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=duplicate-code,line-too-long
# pyright: reportUnknownMemberType=false, reportAttributeAccessIssue=false
# pyright: reportUnknownArgumentType=false, reportOptionalMemberAccess=false
# pyright: reportDeprecated=false, reportMissingTypeStubs=false
# pyright: reportUnsafeMultipleInheritance=false,reportArgumentType=false
# flake8: noqa: E501

"""Step-by-step Waldiez runner with user interaction capabilities."""

import asyncio
import threading
import traceback
import uuid
from collections import deque
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

from pydantic import ValidationError
from typing_extensions import override

from waldiez.models.waldiez import Waldiez
from waldiez.running.step_by_step.command_handler import CommandHandler
from waldiez.running.step_by_step.events_processor import EventProcessor

from ..base_runner import WaldiezBaseRunner
from ..events_mixin import EventsMixin
from ..exceptions import StopRunningException
from ..results_mixin import WaldiezRunResults
from .breakpoints_mixin import BreakpointsMixin
from .step_by_step_models import (
    VALID_CONTROL_COMMANDS,
    WaldiezDebugConfig,
    WaldiezDebugError,
    WaldiezDebugEventInfo,
    WaldiezDebugInputRequest,
    WaldiezDebugInputResponse,
    WaldiezDebugMessage,
    WaldiezDebugStats,
    WaldiezDebugStepAction,
)

if TYPE_CHECKING:
    from autogen.agentchat import ConversableAgent  # type: ignore
    from autogen.events import BaseEvent  # type: ignore
    from autogen.messages import BaseMessage  # type: ignore

MESSAGES = {
    "workflow_starting": "<Waldiez step-by-step> - Starting workflow...",
    "workflow_finished": "<Waldiez step-by-step> - Workflow finished",
    "workflow_stopped": "<Waldiez step-by-step> - Workflow stopped by user",
    "workflow_failed": (
        "<Waldiez step-by-step> - Workflow execution failed: {error}"
    ),
}
DEBUG_INPUT_PROMPT = (
    # cspell: disable-next-line
    "[Step] (c)ontinue, (r)un, (q)uit, (i)nfo, (h)elp, (st)ats: "
)


def gen_id() -> str:
    """Generate a new id.

    Returns
    -------
    str
        The new id.
    """
    return str(uuid.uuid4())


# pylint: disable=too-many-instance-attributes
# noinspection DuplicatedCode,StrFormat
class WaldiezStepByStepRunner(WaldiezBaseRunner, BreakpointsMixin):
    """Refactored step-by-step runner with improved architecture."""

    def __init__(
        self,
        waldiez: Waldiez,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool = False,
        dot_env: str | Path | None = None,
        auto_continue: bool = False,
        breakpoints: Iterable[Any] | None = None,
        config: WaldiezDebugConfig | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the step-by-step runner."""
        super().__init__(
            waldiez,
            output_path=output_path,
            uploads_root=uploads_root,
            structured_io=structured_io,
            dot_env=dot_env,
            **kwargs,
        )
        BreakpointsMixin.__init__(self, config=config)
        self.set_agent_id_to_name(waldiez.flow.unique_names["agent_names"])

        # Configuration
        self._config = config or WaldiezDebugConfig()
        self._config.auto_continue = auto_continue

        # Core state
        self._event_count = 0
        self._processed_events = 0
        self._step_mode = self._config.step_mode

        # Use deque for efficient FIFO operations on event history
        self._event_history: deque[dict[str, Any]] = deque(
            maxlen=self._config.max_event_history
        )
        self._current_event: Union["BaseEvent", "BaseMessage", None] = None

        # Participant tracking
        self._known_participants = self.waldiez.info.participants
        self._last_sender: str | None = None
        self._last_recipient: str | None = None

        # Initialize breakpoints
        if breakpoints:
            _, errors = self.import_breakpoints(list(breakpoints))
            if errors:
                for error in errors:
                    self.log.warning("Breakpoint import error: %s", error)

        # Command handling
        self._command_handler = CommandHandler(self)
        self._event_processor = EventProcessor(self)
        auto_run = self.is_auto_run()
        self._config.auto_continue = auto_run

    @property
    def auto_continue(self) -> bool:
        """Get whether auto-continue is enabled."""
        return self._config.auto_continue

    @auto_continue.setter
    def auto_continue(self, value: bool) -> None:
        """Set whether auto-continue is enabled.

        Parameters
        ----------
        value : bool
            Whether to enable auto-continue.
        """
        self._config.auto_continue = value
        self.log.debug("Auto-continue mode set to: %s", value)

    @property
    def step_mode(self) -> bool:
        """Get the step mode.

        Returns
        -------
        bool
            Whether the step mode is enabled.
        """
        return self._step_mode

    @step_mode.setter
    def step_mode(self, value: bool) -> None:
        """Set the step mode.

        Parameters
        ----------
        value : bool
            Whether to enable step mode.
        """
        self._step_mode = value
        self.log.debug("Step mode set to: %s", value)

    @property
    def last_sender(self) -> str | None:
        """Get the last sender.

        Returns
        -------
        str | None
            The last sender, if available.
        """
        return self._last_sender

    @last_sender.setter
    def last_sender(self, value: str | None) -> None:
        """Set the last sender.

        Parameters
        ----------
        value : str | None
            The last sender to set.
        """
        self._last_sender = value

    @property
    def last_recipient(self) -> str | None:
        """Get the last recipient.

        Returns
        -------
        str | None
            The last recipient, if available.
        """
        return self._last_recipient

    @last_recipient.setter
    def last_recipient(self, value: str | None) -> None:
        """Set the last recipient.

        Parameters
        ----------
        value : str | None
            The last recipient to set.
        """
        self._last_recipient = value

    @property
    def stop_requested(self) -> threading.Event:
        """Get the stop requested event."""
        return self._stop_requested

    @property
    def max_event_history(self) -> int:
        """Get the maximum event history size."""
        return self._config.max_event_history

    def add_to_history(self, event_info: dict[str, Any]) -> None:
        """Add an event to the history.

        Parameters
        ----------
        event_info : dict[str, Any]
            The event information to add to the history.
        """
        self._event_history.append(event_info)

    def pop_event(self) -> None:
        """Pop event from the history."""
        if self._event_history:
            self._event_history.popleft()

    def emit_event(
        self, event: Union["BaseEvent", "BaseMessage", dict[str, Any]]
    ) -> None:
        """Emit an event.

        Parameters
        ----------
        event : BaseEvent | BaseMessage | dict[str, Any]
            The event to emit.
        """
        if not isinstance(event, dict):
            event_info = event.model_dump(
                mode="json", exclude_none=True, fallback=str
            )
            event_info["count"] = self._event_count
            event_info["sender"] = getattr(event, "sender", self._last_sender)
            event_info["recipient"] = getattr(
                event, "recipient", self._last_recipient
            )
        else:
            event_info = event
        self.emit(WaldiezDebugEventInfo(event=event_info))

    # noinspection PyTypeHints
    @override
    def emit(self, message: WaldiezDebugMessage) -> None:
        """Emit a debug message.

        Parameters
        ----------
        message : WaldiezDebugMessage
            The message to emit.
        """
        message_dump = message.model_dump(
            mode="json", exclude_none=True, fallback=str
        )
        self.print(message_dump)

    @property
    def current_event(self) -> Union["BaseEvent", "BaseMessage", None]:
        """Get the current event.

        Returns
        -------
        Union["BaseEvent", "BaseMessage", None]
            The current event, if available.
        """
        return self._current_event

    @current_event.setter
    def current_event(
        self, value: Union["BaseEvent", "BaseMessage", None]
    ) -> None:
        """Set the current event.

        Parameters
        ----------
        value : Union["BaseEvent", "BaseMessage", None]
            The event to set as the current event.
        """
        self._current_event = value

    @property
    def event_count(self) -> int:
        """Get the current event count.

        Returns
        -------
        int
            The current event count.
        """
        return self._event_count

    def event_plus_one(self) -> None:
        """Increment the current event count."""
        self._event_count += 1

    def show_event_info(self) -> None:
        """Show detailed information about the current event."""
        if not self._current_event:
            self.emit(WaldiezDebugError(error="No current event to display"))
            return

        event_info = self._current_event.model_dump(
            mode="json", exclude_none=True, fallback=str
        )
        # Add additional context
        event_info["_meta"] = {
            "event_number": self._event_count,
            "processed_events": self._processed_events,
            "step_mode": self._step_mode,
            "has_breakpoints": len(self._breakpoints) > 0,
        }
        self.emit(WaldiezDebugEventInfo(event=event_info))

    def show_stats(self) -> None:
        """Show comprehensive execution statistics."""
        base_stats: dict[str, Any] = {
            "execution": {
                "events_processed": self._processed_events,
                "total_events": self._event_count,
                "processing_rate": (
                    f"{(self._processed_events / self._event_count * 100):.1f}%"
                    if self._event_count > 0
                    else "0%"
                ),
            },
            "mode": {
                "step_mode": self._step_mode,
                "auto_continue": self._config.auto_continue,
            },
            "history": {
                "event_history_count": len(self._event_history),
                "max_history_size": self._config.max_event_history,
                "memory_usage": f"{len(self._event_history) * 200}B (est.)",
            },
            "participants": {
                "last_sender": self._last_sender,
                "last_recipient": self._last_recipient,
                "known_participants": len(self._known_participants),
            },
        }

        # Merge with breakpoint stats
        breakpoint_stats = self.get_breakpoint_stats()
        stats_dict: dict[str, Any] = {
            **base_stats,
            "breakpoints": breakpoint_stats,
        }

        self.emit(WaldiezDebugStats(stats=stats_dict))

    @property
    def execution_stats(self) -> dict[str, Any]:
        """Get comprehensive execution statistics.

        Returns
        -------
        dict[str, Any]
            A dictionary containing execution statistics.
        """
        base_stats: dict[str, Any] = {
            "total_events": self._event_count,
            "processed_events": self._processed_events,
            "event_processing_rate": (
                self._processed_events / self._event_count
                if self._event_count > 0
                else 0
            ),
            "step_mode": self._step_mode,
            "auto_continue": self._config.auto_continue,
            "event_history_count": len(self._event_history),
            "last_sender": self._last_sender,
            "last_recipient": self._last_recipient,
            "known_participants": [
                p.model_dump() for p in self._known_participants
            ],
            "config": self._config.model_dump(),
        }

        return {**base_stats, "breakpoints": self.get_breakpoint_stats()}

    @property
    def event_history(self) -> list[dict[str, Any]]:
        """Get the history of processed events.

        Returns
        -------
        list[dict[str, Any]]
            A list of dictionaries containing event history.
        """
        return list(self._event_history)

    def reset_session(self) -> None:
        """Reset the debugging session state."""
        self._event_count = 0
        self._processed_events = 0
        self._event_history.clear()
        self._current_event = None
        self._last_sender = None
        self._last_recipient = None
        self.reset_stats()
        self.log.info("Debug session reset")

    def _get_user_response(
        self,
        user_response: str,
        request_id: str,
        skip_id_check: bool = False,
    ) -> tuple[str | None, bool]:
        """Get and validate user response."""
        try:
            response = WaldiezDebugInputResponse.model_validate_json(
                user_response
            )
        except ValidationError as exc:
            # Handle raw CLI input
            got = user_response.strip().lower()
            if got in VALID_CONTROL_COMMANDS:
                return got, True
            self.emit(WaldiezDebugError(error=f"Invalid input: {exc}"))
            return None, False

        if not skip_id_check and response.request_id != request_id:
            self.emit(
                WaldiezDebugError(
                    error=f"Stale input received: {response.request_id} != {request_id}"
                )
            )
            return None, False

        return response.data, True

    def _parse_user_action(
        self, user_response: str, request_id: str
    ) -> WaldiezDebugStepAction:
        """Parse user action using the command handler."""
        self.log.debug("Parsing user action... '%s'", user_response)

        user_input, is_valid = self._get_user_response(
            user_response,
            request_id=request_id,
            skip_id_check=True,
        )
        if not is_valid:
            return WaldiezDebugStepAction.UNKNOWN

        return self._command_handler.handle_command(user_input or "")

    def _get_user_action(self, force: bool) -> WaldiezDebugStepAction:
        """Get user action with timeout support.

        Parameters
        ----------
        force : bool
            Force getting the user's action, even if in auto-run mode.
        """
        if self._config.auto_continue:
            self.step_mode = True
            if force:
                self._config.auto_continue = False
            else:
                return WaldiezDebugStepAction.CONTINUE
        while True:
            request_id = gen_id()
            try:
                if not self.structured_io:
                    # if structured, we already do this (print the prompt)
                    self.emit(
                        WaldiezDebugInputRequest(
                            prompt=DEBUG_INPUT_PROMPT, request_id=request_id
                        )
                    )
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.log.warning("Failed to emit input request: %s", e)
            try:
                user_input = EventsMixin.get_user_input(
                    DEBUG_INPUT_PROMPT,
                    request_id=request_id,
                ).strip()
                return self._parse_user_action(
                    user_input, request_id=request_id
                )

            except (KeyboardInterrupt, EOFError):
                self._stop_requested.set()
                return WaldiezDebugStepAction.QUIT

    async def _a_get_user_action(self, force: bool) -> WaldiezDebugStepAction:
        """Get user action asynchronously."""
        if self._config.auto_continue:
            self.step_mode = True
            if force:
                self._config.auto_continue = False
            else:
                return WaldiezDebugStepAction.CONTINUE

        while True:
            request_id = gen_id()
            # pylint: disable=too-many-try-statements
            try:
                self.emit(
                    WaldiezDebugInputRequest(
                        prompt=DEBUG_INPUT_PROMPT, request_id=request_id
                    )
                )

                user_input = await EventsMixin.a_get_user_input(
                    DEBUG_INPUT_PROMPT
                )
                user_input = user_input.strip()
                return self._parse_user_action(
                    user_input, request_id=request_id
                )

            except (KeyboardInterrupt, EOFError):
                return WaldiezDebugStepAction.QUIT

    def _handle_step_interaction(self, force: bool) -> bool:
        """Handle step-by-step user interaction.

        Parameters
        ----------
        force : bool
            Force getting the user's action, even if in auto-run mode.
        """
        while True:
            action = self._get_user_action(force)
            if action in (
                WaldiezDebugStepAction.CONTINUE,
                WaldiezDebugStepAction.STEP,
            ):
                return True
            if action == WaldiezDebugStepAction.RUN:
                self._config.auto_continue = True
                return True
            if action == WaldiezDebugStepAction.QUIT:
                return False
            # For other actions (info, help, etc.), continue the loop

    async def _a_handle_step_interaction(self, force: bool) -> bool:
        """Handle step-by-step user interaction asynchronously."""
        while True:
            action = await self._a_get_user_action(force)
            if action in (
                WaldiezDebugStepAction.CONTINUE,
                WaldiezDebugStepAction.STEP,
            ):
                return True
            if action == WaldiezDebugStepAction.RUN:
                self._config.auto_continue = True
                return True
            if action == WaldiezDebugStepAction.QUIT:
                return False
            # For other actions (info, help, etc.), continue the loop

    @override
    def _run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez workflow with step-by-step debugging."""
        # pylint: disable=import-outside-toplevel
        from autogen.io import IOStream  # type: ignore

        from waldiez.io import StructuredIOStream

        results_container: WaldiezRunResults = {
            "results": [],
            "exception": None,
            "completed": False,
        }
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            loaded_module = self._load_module(output_file, temp_dir)
            WaldiezBaseRunner._store_run_paths(
                tmp_dir=temp_dir, output_file=output_file
            )
            if self._stop_requested.is_set():
                self.log.debug(
                    "Step-by-step execution stopped before workflow start"
                )
                return []

            # Setup I/O
            if self.structured_io:
                stream = StructuredIOStream(
                    uploads_root=uploads_root, is_async=False
                )
            else:
                stream = IOStream.get_default()

            EventsMixin.set_print_function(stream.print)
            EventsMixin.set_input_function(stream.input)
            EventsMixin.set_send_function(stream.send)
            self._output_dir = temp_dir
            self.print(MESSAGES["workflow_starting"])
            self.print(self.waldiez.info.model_dump_json())
            results = loaded_module.main(
                on_event=self._on_event,
                state_json=self.state_json,
            )
            results_container["results"] = results
            self.print(MESSAGES["workflow_finished"])

        except Exception as e:
            if StopRunningException.reason in str(e):
                raise StopRunningException(StopRunningException.reason) from e
            results_container["exception"] = e
            traceback.print_exc()
            self.print(MESSAGES["workflow_failed"].format(error=str(e)))
        finally:
            results_container["completed"] = True
            WaldiezBaseRunner._remove_run_paths()

        return results_container["results"]

    def _re_emit_if_needed(self, event_info: dict[str, Any]) -> None:
        # emit again if type is text, swapping the sender and without recipient
        if event_info.get("type", "") == "text":
            event_info["sender"] = event_info["recipient"]
            event_info["recipient"] = None
            event_info["agents"]["sender"] = event_info["agents"]["recipient"]
            event_info["agents"]["recipient"] = None
            self.emit_event(event_info)

    def _on_event(
        self,
        event: Union["BaseEvent", "BaseMessage"],
        agents: list["ConversableAgent"],
    ) -> bool:
        """Process an event with step-by-step debugging."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            # Use the event processor for core logic
            result = self._event_processor.process_event(event, agents)

            if result["action"] == "stop":
                self.log.debug(
                    "Step-by-step execution stopped before event processing"
                )
                return False
            event_info = result["event_info"]
            self.emit_event(event_info)
            # Handle breakpoint logic
            if result["action"] == "break":
                if not self._handle_step_interaction(force=True):
                    self._stop_requested.set()
                    if hasattr(event, "type") and event.type == "input_request":
                        event.content.respond("exit")
                        return True
                    raise StopRunningException(StopRunningException.reason)
                self._re_emit_if_needed(event_info)
            # Process the actual event
            EventsMixin.process_event(
                event, agents, output_dir=self._output_dir, skip_send=True
            )
            self._processed_events += 1

        except Exception as e:
            if not isinstance(e, StopRunningException):
                raise RuntimeError(
                    f"Error processing event {event}: {e}\n{traceback.format_exc()}"
                ) from e
            raise StopRunningException(StopRunningException.reason) from e

        return not self._stop_requested.is_set()

    # pylint: disable=too-complex
    @override
    async def _a_run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez workflow with step-by-step debugging (async)."""

        async def _execute_workflow() -> list[dict[str, Any]]:
            # pylint: disable=import-outside-toplevel
            from autogen.io import IOStream

            from waldiez.io import StructuredIOStream

            # pylint: disable=too-many-try-statements,broad-exception-caught
            try:
                loaded_module = self._load_module(output_file, temp_dir)
                await WaldiezBaseRunner._a_store_run_paths(
                    tmp_dir=temp_dir, output_file=output_file
                )
                if self._stop_requested.is_set():
                    self.log.debug(
                        "Step-by-step execution stopped before workflow start"
                    )
                    return []

                if self.structured_io:
                    stream = StructuredIOStream(
                        uploads_root=uploads_root, is_async=True
                    )
                else:
                    stream = IOStream.get_default()

                EventsMixin.set_print_function(stream.print)
                EventsMixin.set_input_function(stream.input)
                EventsMixin.set_send_function(stream.send)

                self._output_dir = temp_dir
                self.print(MESSAGES["workflow_starting"])
                self.print(self.waldiez.info.model_dump_json())

                results = await loaded_module.main(
                    on_event=self._a_on_event,
                    state_json=self.state_json,
                )
                self.print(MESSAGES["workflow_finished"])
                return results

            except Exception as e:
                if StopRunningException.reason in str(e):
                    raise StopRunningException(
                        StopRunningException.reason
                    ) from e
                self.print(MESSAGES["workflow_failed"].format(error=str(e)))
                traceback.print_exc()
                return []
            finally:
                WaldiezBaseRunner._remove_run_paths()

        # Create and monitor cancellable task
        task = asyncio.create_task(_execute_workflow())
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            while not task.done():
                if self._stop_requested.is_set():
                    task.cancel()
                    self.log.debug("Step-by-step execution stopped by user")
                    break
                await asyncio.sleep(0.1)
            return await task
        except asyncio.CancelledError:
            self.log.debug("Step-by-step execution cancelled")
            return []

    async def _a_on_event(
        self,
        event: Union["BaseEvent", "BaseMessage"],
        agents: list["ConversableAgent"],
    ) -> bool:
        """Process an event with step-by-step debugging asynchronously."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            # Use the event processor for core logic
            result = self._event_processor.process_event(event, agents)

            if result["action"] == "stop":
                self.log.debug(
                    "Async step-by-step execution stopped before event processing"
                )
                return False
            event_info = result["event_info"]
            self.emit_event(event_info)
            # Handle breakpoint logic
            if result["action"] == "break":
                if not await self._a_handle_step_interaction(force=True):
                    self._stop_requested.set()
                    if hasattr(event, "type") and event.type == "input_request":
                        await event.content.respond("exit")
                        return True
                    raise StopRunningException(StopRunningException.reason)
                self._re_emit_if_needed(event_info)
            # Process the actual event
            await EventsMixin.a_process_event(
                event, agents, output_dir=self._output_dir, skip_send=True
            )
            self._processed_events += 1

        except Exception as e:
            if not isinstance(e, StopRunningException):
                raise RuntimeError(
                    f"Error processing event {event}: {e}\n{traceback.format_exc()}"
                ) from e
            raise StopRunningException(StopRunningException.reason) from e

        return not self._stop_requested.is_set()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=unused-argument
# pyright: reportUnusedParameter=false
"""Image media content models."""

from pathlib import Path

from pydantic import BaseModel
from typing_extensions import Literal

from ...utils import get_image
from .base import ImageContent


# noinspection DuplicatedCode
class ImageMediaContent(BaseModel):
    """Image media content."""

    type: Literal["image"] = "image"
    image: ImageContent

    def to_string(
        self,
        uploads_root: Path | None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        if self.image.url:
            image = get_image(
                uploads_root=uploads_root,
                image_data=self.image.url,
                base_name=base_name,
            )
            return f"<img {image}>"
        if self.image.file:
            image = get_image(
                uploads_root=uploads_root,
                image_data=self.image.file,
                base_name=base_name,
            )
            return f"<img {image}>"
        return str(self.image.file)


class ImageUrlMediaContent(BaseModel):
    """Image URL media content."""

    type: Literal["image_url"] = "image_url"
    image_url: ImageContent

    # noinspection DuplicatedCode
    def to_string(
        self,
        uploads_root: Path | None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        if self.image_url.url:
            image = get_image(
                uploads_root=uploads_root,
                image_data=self.image_url.url,
                base_name=base_name,
            )
            return f"<img {image}>"
        if self.image_url.file:
            image = get_image(
                uploads_root=uploads_root,
                image_data=self.image_url.file,
                base_name=base_name,
            )
            return f"<img {image}>"
        return str(self.image_url.file)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Tool related models."""

from .extra_requirements import get_tools_extra_requirements
from .tool import SHARED_TOOL_NAME, WaldiezTool
from .tool_data import WaldiezToolData
from .tool_type import WaldiezToolType

__all__ = [
    "SHARED_TOOL_NAME",
    "WaldiezTool",
    "WaldiezToolData",
    "WaldiezToolType",
    "get_tools_extra_requirements",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez agent connection model."""

from typing import TypedDict

from ..agents import (
    WaldiezAgent,
)
from ..chat import WaldiezChat


class WaldiezAgentConnection(TypedDict):
    """Agent connection."""

    source: WaldiezAgent
    target: WaldiezAgent
    chat: WaldiezChat



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG related extras."""

from dataclasses import dataclass, field

from waldiez.exporting.core.result import ExportResult

from ...enums import AgentPosition, ExportPosition
from .standard_extras import StandardExtras


@dataclass
class RAGUserExtras(StandardExtras):
    """RAG configuration."""

    before_content: str = ""
    imports: set[str] = field(default_factory=set)

    def _contribute_specific_content(self, result: ExportResult) -> None:
        """Contribute RAG specific content to the export result."""
        if self.extra_args:
            for arg in self.extra_args:
                result.add_instance_argument(
                    name=arg.name,
                    value=arg.value,
                    instance_id=arg.instance_id,
                    tabs=arg.tabs,
                )
        if self.imports:
            for imp in self.imports:
                result.add_import(imp)
        if self.before_content.strip():
            result.add_content(
                self.before_content,
                position=ExportPosition.AGENTS,
                agent_position=AgentPosition.BEFORE,
                agent_id=self.instance_id,
            )

    def has_specific_content(self) -> bool:
        """Check if there's any RAG content.

        Returns
        -------
        bool
            True if there's any RAG configuration.
        """
        if not super().has_specific_content():
            return bool(
                self.extra_args or self.before_content.strip() or self.imports
            )
        return True



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=line-too-long
# flake8: noqa: E501
"""Predefined Google search tool for Waldiez."""

import os
from typing import Any

from ._config import PredefinedToolConfig
from .protocol import PredefinedTool


class GoogleSearchToolImpl(PredefinedTool):
    """Google search tool for Waldiez."""

    required_secrets: list[str] = [
        "GOOGLE_SEARCH_API_KEY",
        "GOOGLE_SEARCH_ENGINE_ID",
    ]
    required_kwargs: dict[str, type] = {}
    _kwargs: dict[str, Any] = {}

    @property
    def name(self) -> str:
        """Tool name."""
        return "google_search"

    @property
    def description(self) -> str:
        """Tool description."""
        return "Search Google for a given query."

    @property
    def kwargs(self) -> dict[str, Any]:
        """Keyword arguments for the tool, used for initialization."""
        return self._kwargs

    @property
    def requirements(self) -> list[str]:
        """Python requirements for the tool."""
        return ["ag2[google-search,gemini]"]

    @property
    def tags(self) -> list[str]:
        """Tags for the tool, used for categorization."""
        return ["google", "search", "web"]

    @property
    def tool_imports(self) -> list[str]:
        """Imports required for the tool implementation."""
        return [
            "from autogen.tools.experimental import GoogleSearchTool",
        ]

    @property
    def google_search_engine_id(self) -> str:
        """Google search engine ID."""
        from_env = os.environ.get("GOOGLE_SEARCH_ENGINE_ID", "")
        if from_env:
            return from_env
        return str(self.kwargs.get("google_search_engine_id", ""))

    def validate_secrets(self, secrets: dict[str, str]) -> list[str]:
        """Validate secrets and return list of missing required ones.

        Parameters
        ----------
        secrets : dict[str, str]
            Dictionary of secrets/environment variables.

        Returns
        -------
        list[str]
            List of missing required secrets.
        """
        missing: list[str] = []
        for secret in self.required_secrets:
            if secret not in secrets:
                missing.append(secret)
        return missing

    def validate_kwargs(self, kwargs: dict[str, str]) -> list[str]:
        """Validate keyword arguments and return list of missing required ones.

        Parameters
        ----------
        kwargs : dict[str, str]
            Dictionary of keyword arguments.

        Returns
        -------
        list[str]
            List of missing required keyword arguments.

        Raises
        ------
        ValueError
            If any keyword argument is invalid
            (cannot be cast to the required type).
        """
        missing: list[str] = []
        invalid: list[str] = []
        for name, type_of in self.required_kwargs.items():
            if name not in kwargs:
                missing.append(name)
            elif not isinstance(kwargs[name], type_of):
                # pylint: disable=broad-exception-caught
                # noinspection PyBroadException
                try:
                    kwargs[name] = type_of(kwargs[name])
                    self._kwargs[name] = kwargs[name]
                except Exception:
                    invalid.append(name)
                else:
                    if name in self._kwargs:
                        self._kwargs[name] = kwargs[name]
            else:
                self._kwargs[name] = kwargs[name]
        if invalid:
            raise ValueError(f"Invalid keyword arguments: {invalid}")
        return missing

    def get_content(
        self,
        secrets: dict[str, str],
        runtime_kwargs: dict[str, Any] | None = None,
    ) -> str:
        """Get content for the tool.

        Parameters
        ----------
        secrets : dict[str, str]
            Dictionary of secrets/environment variables.
        runtime_kwargs : dict[str, Any] | None, optional
            Runtime keyword arguments to customize the content generation.

        Returns
        -------
        str
            The content for the tool.
        """
        content = f'''
def {self.name}(
    query: str,
    num_results: int = 10,
) -> list[dict[str, Any]]:
    """Perform a Google search and return formatted results.

    Args:
        query: The search query string.
        num_results: The maximum number of results to return. Defaults to 10.
    Returns:
        A list of dictionaries of the search results.
    """
    google_search_api_key = os.environ.get("GOOGLE_SEARCH_API_KEY", "")
    if not google_search_api_key:
        raise ValueError("GOOGLE_SEARCH_API_KEY is required for Google search tool.")
    google_search_engine_id = os.environ.get("GOOGLE_SEARCH_ENGINE_ID", "")
    if not google_search_engine_id:
        raise ValueError("Google Search Engine ID is required for Google search tool.")
    {self.name}_tool = GoogleSearchTool(
        search_api_key=google_search_api_key,
        search_engine_id=google_search_engine_id,
    )
    return {self.name}_tool(
        query=query,
        search_api_key=google_search_api_key,
        search_engine_id=google_search_engine_id,
        num_results=num_results
    )
'''
        return content


# pylint: disable=invalid-name
GoogleSearchTool = GoogleSearchToolImpl()
GoogleSearchConfig = PredefinedToolConfig(
    name=GoogleSearchTool.name,
    description=GoogleSearchTool.description,
    required_secrets=GoogleSearchTool.required_secrets,
    required_kwargs=GoogleSearchTool.required_kwargs,
    requirements=GoogleSearchTool.requirements,
    tags=GoogleSearchTool.tags,
    implementation=GoogleSearchTool,
)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get pgvector related content and imports."""

from waldiez.models import WaldiezRagUserProxy


def _get_pgvector_client_string(agent: WaldiezRagUserProxy) -> tuple[str, str]:
    """Get the PGVectorDB client string.

    Parameters
    ----------
    agent : WaldiezRagUserProxy
        The agent.

    Returns
    -------
    tuple[str, str]
        The 'client' and what to import.
    """
    to_import = "import psycopg"
    client_str = "psycopg."
    connection_url = agent.retrieve_config.db_config.connection_url
    client_str += f'connect("{connection_url}")'
    return client_str, to_import


def _get_pgvector_embedding_function_string(
    agent: WaldiezRagUserProxy, agent_name: str
) -> tuple[str, str, str]:
    """Get the PGVectorDB embedding function string.

    Parameters
    ----------
    agent : WaldiezRagUserProxy
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    tuple[str, str, str]
        The 'embedding_function', the import and the custom_embedding_function.
    """
    to_import = ""
    embedding_function_content = ""
    if agent.retrieve_config.use_custom_embedding:
        embedding_function_content, embedding_function_arg = (
            agent.retrieve_config.get_custom_embedding_function(
                name_suffix=agent_name
            )
        )
        embedding_function_content = "\n" + embedding_function_content
    else:
        to_import = "from sentence_transformers import SentenceTransformer"
        embedding_function_arg = "SentenceTransformer("
        embedding_function_arg += (
            f'"{agent.retrieve_config.db_config.model}").encode'
        )
    return embedding_function_arg, to_import, embedding_function_content


def get_pgvector_db_args(
    agent: WaldiezRagUserProxy, agent_name: str
) -> tuple[str, set[str], str]:
    """Get the kwargs to use for PGVectorDB.

    Parameters
    ----------
    agent : WaldiezRagUserProxy
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    tuple[tuple[str,str], Set[str], str]
        The kwargs to use, what to import and the custom_embedding_function.
    """
    client_str, to_import_client = _get_pgvector_client_string(agent)
    embedding_function_arg, to_import_embedding, embedding_function_body = (
        _get_pgvector_embedding_function_string(agent, agent_name)
    )
    to_import = (
        {to_import_client, to_import_embedding}
        if to_import_embedding
        else {to_import_client}
    )
    kwarg_str = (
        f"            client={client_str},"
        "\n"
        f"            embedding_function={embedding_function_arg},"
        "\n"
    )
    return kwarg_str, to_import, embedding_function_body



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Tool model."""

from typing import Any

from pydantic import Field, SerializationInfo, model_serializer, model_validator
from typing_extensions import Annotated, Self

from ..common import WaldiezBase
from .tool_type import WaldiezToolType


class WaldiezToolData(WaldiezBase):
    """Waldiez Tool Data.

    Attributes
    ----------
    tool_type : WaldiezToolType
        The type of the tool: shared, custom, langchain, crewai.
    content : str
        The content (source code) of the tool.
    secrets : dict[str, str]
        The secrets (environment variables) of the tool.
    """

    tool_type: Annotated[
        WaldiezToolType,
        Field(
            "custom",
            alias="toolType",
            title="Tool Type",
            description=(
                "The type of the tool: shared, custom, langchain, crewai."
            ),
        ),
    ]
    content: Annotated[
        str,
        Field(
            ...,
            title="Content",
            description="The content (source code) of the tool.",
        ),
    ]
    secrets: Annotated[
        dict[str, str],
        Field(
            default_factory=dict,
            title="Secrets",
            description="The secrets (environment variables) of the tool.",
        ),
    ]
    kwargs: Annotated[
        dict[str, Any],
        Field(
            default_factory=dict,
            title="Keyword Arguments",
            description=(
                "Keyword arguments for the tool, used for initialization."
            ),
        ),
    ]

    _raw_content: str = ""

    @model_validator(mode="after")
    def validate_tool_data(self) -> Self:
        """Validate the tool data.

        Returns
        -------
        Self
            The validated tool data.
        """
        self._raw_content = self.content

        return self

    @model_serializer(mode="plain", when_used="always")
    def serialize_tool_data(self, info: SerializationInfo) -> dict[str, Any]:
        """Serialize the tool data.

        Parameters
        ----------
        info : SerializationInfo
            The serialization information.

        Returns
        -------
        dict[str, Any]
            The serialized tool data.
        """
        tool_type_key = "toolType" if info.by_alias else "tool_type"
        return {
            tool_type_key: self.tool_type,
            "content": self._raw_content,
            "secrets": self.secrets,
            "kwargs": self.kwargs,
        }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Manages workflow sessions across WebSocket clients."""

import asyncio
import logging
import time
from collections import defaultdict
from pathlib import Path
from typing import Any, final

from .models import ExecutionMode, SessionState, WorkflowStatus
from .session import WorkflowSession
from .session_stats import SessionStats


# noinspection TryExceptPass,PyBroadException
@final
class SessionManager:
    """Manage workflow sessions across WebSocket clients."""

    def __init__(
        self,
        cleanup_interval: float = 300.0,
        max_session_age: float = 3600.0,
    ) -> None:
        """Initialize the session manager.

        Parameters
        ----------
        cleanup_interval : float
            The interval at which to clean up expired sessions
        max_session_age : float
            The maximum age of a session before it is considered expired
        """
        self._sessions: dict[str, WorkflowSession] = {}
        self._client_sessions: dict[str, list[str]] = defaultdict(list)
        self._stats = SessionStats()
        self._cleanup_interval = cleanup_interval
        self._max_session_age = max_session_age
        self._cleanup_task: asyncio.Task[Any] | None = None
        self._lock = asyncio.Lock()
        self._stop_event = asyncio.Event()
        self._logger = logging.getLogger(__name__)

    # ---------------- lifecycle ----------------

    async def start(self) -> None:
        """Start the session manager."""
        self._stop_event.clear()
        if self._cleanup_task is None:
            self._cleanup_task = asyncio.create_task(self._cleanup_loop())

    async def stop(self) -> None:
        """Stop the session manager."""
        self._stop_event.set()
        if self._cleanup_task:
            self._cleanup_task.cancel()
            try:
                await self._cleanup_task
            except asyncio.CancelledError:
                pass
            self._cleanup_task = None
        await self.cleanup_all_sessions()

    # ---------------- session ops ----------------

    async def create_session(
        self,
        session_id: str,
        client_id: str,
        mode: ExecutionMode,
        runner: Any = None,
        temp_file: Path | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> WorkflowSession:
        """Create a new workflow session.

        Parameters
        ----------
        session_id : str
            The ID of the session to create
        client_id : str
            The ID of the client creating the session
        mode : ExecutionMode
            The execution mode for the session
        runner : Any, optional
            The runner to use for the session
        temp_file : Path | None, optional
            The temporary file to use for the session
        metadata : dict[str, Any] | None, optional
            Metadata to associate with the session

        Returns
        -------
        WorkflowSession
            The created workflow session

        Raises
        ------
        ValueError
            If a session with the given ID already exists
        """
        session_state = SessionState(
            session_id=session_id,
            client_id=client_id,
            status=WorkflowStatus.IDLE,
            mode=mode,
            metadata=metadata or {},
        )
        session = WorkflowSession(
            session_state=session_state, runner=runner, temp_file=temp_file
        )

        async with self._lock:
            if session_id in self._sessions:
                raise ValueError(f"Session {session_id} already exists")
            self._sessions[session_id] = session
            self._client_sessions[client_id].append(session_id)
            self._recompute_stats_locked()
        return session

    async def get_session(self, session_id: str) -> WorkflowSession | None:
        """Get a workflow session by ID.

        Parameters
        ----------
        session_id : str
            The ID of the session to retrieve

        Returns
        -------
        WorkflowSession | None
            The workflow session with the given ID, or None if it does not exist
        """
        async with self._lock:
            return self._sessions.get(session_id)

    async def get_client_sessions(
        self, client_id: str
    ) -> list[WorkflowSession]:
        """Get all workflow sessions for a client.

        Parameters
        ----------
        client_id : str
            The ID of the client to retrieve sessions for

        Returns
        -------
        list[WorkflowSession]
            A list of workflow sessions for the client
        """
        async with self._lock:
            ids = list(self._client_sessions.get(client_id, []))
            return [self._sessions[sid] for sid in ids if sid in self._sessions]

    async def update_session_status(
        self,
        session_id: str,
        new_status: WorkflowStatus,
    ) -> bool:
        """Update the status of a workflow session.

        Parameters
        ----------
        session_id : str
            The ID of the session to update
        new_status : WorkflowStatus
            The new status to set for the session

        Returns
        -------
        bool
            True if the status was updated, False if the session was not found
        """
        async with self._lock:
            session = self._sessions.get(session_id)
            if not session:
                return False
            session.update_status(new_status)
            self._recompute_stats_locked()
            return True

    async def get_session_mode(self, session_id: str) -> ExecutionMode | None:
        """Get the execution mode of a workflow session.

        Parameters
        ----------
        session_id : str
            The ID of the session to retrieve the execution mode for

        Returns
        -------
        ExecutionMode | None
            The execution mode of the session, or None if it does not exist
        """
        async with self._lock:
            session = self._sessions.get(session_id)
            return session.mode if session else None

    async def remove_session(self, session_id: str) -> bool:
        """Remove a workflow session by ID.

        Parameters
        ----------
        session_id : str
            The ID of the session to remove

        Returns
        -------
        bool
            True if the session was removed, False if it did not exist
        """
        self._logger.debug("Removing session %s", session_id)
        # Detach under lock
        async with self._lock:
            session = self._sessions.pop(session_id, None)
            if not session:
                return False
            client_id = session.client_id
            if client_id in self._client_sessions:
                try:
                    self._client_sessions[client_id].remove(session_id)
                    if not self._client_sessions[client_id]:
                        del self._client_sessions[client_id]
                except ValueError:
                    pass
        # Cleanup outside lock
        try:
            session.cleanup()
        except Exception:  # pylint: disable=broad-exception-caught
            pass
        # Update stats under lock
        async with self._lock:
            self._stats.cleanup_count += 1
            self._recompute_stats_locked()
        return True

    async def remove_client_sessions(self, client_id: str) -> int:
        """Remove all workflow sessions for a client.

        Parameters
        ----------
        client_id : str
            The ID of the client whose sessions should be removed

        Returns
        -------
        int
            Number of removed sessions
        """
        async with self._lock:
            sids = list(self._client_sessions.get(client_id, []))
        removed = 0
        for sid in sids:
            if await self.remove_session(sid):
                removed += 1
        return removed

    # ---------------- stats / status ----------------

    async def get_stats(self) -> SessionStats:
        """Get session statistics.

        Returns
        -------
        SessionStats
            Session statistics
        """
        async with self._lock:
            self._recompute_stats_locked()
            return self._stats

    async def get_session_count(self) -> int:
        """Get total number of sessions.

        Returns
        -------
        int
            Total number of sessions
        """
        async with self._lock:
            return len(self._sessions)

    async def get_client_count(self) -> int:
        """Get number of clients with sessions.

        Returns
        -------
        int
            Number of clients with sessions
        """
        async with self._lock:
            return len(self._client_sessions)

    async def get_status(self) -> dict[str, Any]:
        """Get detailed status of the session manager.

        Returns
        -------
        dict[str, Any]
            Detailed status information
        """
        stats = await self.get_stats()
        async with self._lock:
            return {
                "session_manager": {
                    "total_sessions": len(self._sessions),
                    "total_clients": len(self._client_sessions),
                    "cleanup_interval": self._cleanup_interval,
                    "max_session_age": self._max_session_age,
                    "cleanup_task_running": self._cleanup_task is not None
                    and not self._cleanup_task.done(),
                },
                "statistics": stats.model_dump(),
                "timestamp": time.time(),
            }

    # ---------------- cleanup ----------------

    async def cleanup_old_sessions(self, max_age: float | None = None) -> int:
        """Cleanup old sessions.

        Parameters
        ----------
        max_age : float | None
            The maximum age of sessions to clean up

        Returns
        -------
        int
            The number of sessions removed
        """
        max_age = max_age or self._max_session_age
        now_ns = time.monotonic_ns()

        async with self._lock:
            to_remove: list[str] = []
            for sid, session in self._sessions.items():
                remove = False
                if session.raw_state.is_completed:
                    from_now_ns = now_ns - (
                        session.raw_state.end_time
                        or session.raw_state.start_time
                    )
                    age_s = from_now_ns / 1_000_000_000
                    remove = age_s > max_age
                elif not session.raw_state.is_active:
                    age_ns = now_ns - session.last_accessed
                    age_s = age_ns / 1_000_000_000
                    remove = age_s > (max_age * 2)
                if remove:
                    to_remove.append(sid)

        removed = 0
        for sid in to_remove:
            if await self.remove_session(sid):
                removed += 1
        return removed

    async def cleanup_all_sessions(self) -> None:
        """Cleanup all sessions."""
        # Detach under lock
        async with self._lock:
            sessions = list(self._sessions.values())
            self._sessions.clear()
            self._client_sessions.clear()
            self._stats = SessionStats()
        # Cleanup outside lock
        for s in sessions:
            try:
                s.cleanup()
            except Exception:  # pylint: disable=broad-exception-caught
                pass

    async def _cleanup_loop(self) -> None:
        """Cleanup loop for old sessions."""
        while not self._stop_event.is_set():
            try:
                await asyncio.sleep(self._cleanup_interval)
                await self.cleanup_old_sessions()
            except asyncio.CancelledError:
                break
            except Exception:  # pylint: disable=broad-exception-caught
                async with self._lock:
                    self._stats.error_count += 1

    # ---------------- internal ----------------

    def _recompute_stats_locked(self) -> None:
        """Recompute session statistics."""
        self._stats.update_from_sessions(list(self._sessions.values()))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pyright: reportUnknownVariableType=false
# pylint: disable=broad-exception-caught,no-member

"""Session statistics model."""

from pydantic import BaseModel, Field

from .models import WorkflowStatus
from .session import WorkflowSession


class SessionStats(BaseModel):
    """Statistics for session management."""

    total_sessions: int = 0
    active_sessions: int = 0
    completed_sessions: int = 0
    failed_sessions: int = 0
    cancelled_sessions: int = 0

    sessions_by_client: dict[str, int] = Field(default_factory=dict)
    sessions_by_mode: dict[str, int] = Field(default_factory=dict)
    sessions_by_status: dict[str, int] = Field(default_factory=dict)

    average_duration: float = 0.0
    total_duration: float = 0.0

    cleanup_count: int = 0
    error_count: int = 0

    def update_from_sessions(self, sessions: list[WorkflowSession]) -> None:
        """Update stats from current sessions.

        Parameters
        ----------
        sessions : list[WorkflowSession]
            The list of sessions to update stats from.
        """
        # Reset counters
        self.total_sessions = len(sessions)
        self.active_sessions = 0
        self.completed_sessions = 0
        self.failed_sessions = 0
        self.cancelled_sessions = 0

        self.sessions_by_client.clear()
        self.sessions_by_mode.clear()
        self.sessions_by_status.clear()

        total_duration = 0.0
        completed_count = 0

        for session in sessions:
            state = session.state

            if state.status == WorkflowStatus.COMPLETED:
                self.completed_sessions += 1
            elif state.status == WorkflowStatus.FAILED:
                self.failed_sessions += 1
            elif state.status == WorkflowStatus.CANCELLED:
                self.cancelled_sessions += 1
            elif state.is_active:
                self.active_sessions += 1

            client_count = self.sessions_by_client.get(state.client_id, 0)
            self.sessions_by_client[state.client_id] = client_count + 1
            mode_key = state.mode.value
            mode_count = self.sessions_by_mode.get(mode_key, 0)
            self.sessions_by_mode[mode_key] = mode_count + 1
            status_key = state.status.value
            # noinspection PyTypeChecker
            status_count = self.sessions_by_status.get(status_key, 0)
            # noinspection PyTypeChecker
            self.sessions_by_status[status_key] = status_count + 1
            if state.is_completed:
                total_duration += state.duration
                completed_count += 1

        self.total_duration = total_duration
        self.average_duration = (
            total_duration / completed_count if completed_count > 0 else 0.0
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Storage protocol for workspace and checkpoint management."""
# pylint: disable=unnecessary-ellipsis

from collections.abc import Generator, Iterable, Mapping
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, Protocol, runtime_checkable

from typing_extensions import Self

from .checkpoint import WaldiezCheckpoint, WaldiezCheckpointInfo


@runtime_checkable
class Storage(Protocol):  # pragma: no cover
    """Protocol for handling workspace directory and checkpoints management."""

    @property
    def workspace_dir(self) -> Path:
        """Base workspace directory."""
        ...

    @staticmethod
    def load_dict(json_file: Path) -> dict[str, Any]:
        """Load dict from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.
        """
        ...

    @staticmethod
    def load_list(
        json_file: Path, fallback_dict_key: str
    ) -> list[dict[str, Any]]:
        """Load list from json.

        Parameters
        ----------
        json_file : Path
            The path of the file to load data from.
        fallback_dict_key : str
            The key to get the list from if the loaded data is dict.
        """
        ...

    @contextmanager
    def transaction(self) -> Generator[Self, None, None]:
        """Batch multiple operations together.

        Raises
        ------
        Exception
            If any of the operations fails.

        Example
        -------
        with storage.transaction():
            storage.save_checkpoint("session1", state1)
            storage.save_checkpoint("session2", state2)
            storage.delete_checkpoint("session3", timestamp)
        """
        ...

    def save_checkpoint(
        self,
        session_name: str,
        state: dict[str, Any],
        metadata: dict[str, Any] | None = None,
        timestamp: datetime | None = None,
    ) -> Path:
        """Save a checkpoint for a session.

        Parameters
        ----------
        session_name : str
            Name of the session
        state : dict[str, Any]
            State data to save
        metadata : dict[str, Any]
            Optional metadata to include
        timestamp: datetime
            Optional timestamp (defaults to now)
        """
        ...

    def get_checkpoint(
        self, session_name: str, timestamp: datetime | None = None
    ) -> WaldiezCheckpointInfo | None:
        """Load a checkpoint for a session.

        Parameters
        ----------
        session_name : str
            Name of the session
        timestamp: datetime
            Optional timestamp (defaults to latest)

        Raises
        ------
        FileNotFoundError
            If no checkpoint found
        """
        ...

    def load_checkpoint(
        self, info: WaldiezCheckpointInfo, history_index: int | None
    ) -> WaldiezCheckpoint:
        """Load a checkpoint.

        Parameters
        ----------
        info: WaldiezCheckpointInfo
            The checkpoint info to get the path.
        history_index: int | None
            Optional history index to use.
        """
        ...

    def link_checkpoint(
        self,
        to: Path,
        session_name: str,
        timestamp: datetime | None = None,
    ) -> None:
        """Create a symlink to a checkpoint.

        Parameters
        ----------
        to: Path
            Where to create the symlink to.
        session_name : str
            Name of the session
        timestamp: datetime | None
            Optional specific checkpoint timestamp (defaults to latest)

        Raises
        ------
        FileNotFoundError
            If no checkpoint found
        """
        ...

    def list_checkpoints(
        self, session_name: str | None = None
    ) -> list[WaldiezCheckpointInfo]:
        """List available checkpoints.

        Parameters
        ----------
        session_name : str | None
            Optional filter by session name
        """
        ...

    def list_sessions(self) -> list[str]:
        """List available sessions."""
        ...

    def delete_session(self, session_name: str) -> None:
        """Delete a session and all its checkpoints.

        Parameters
        ----------
        session_name : str
            The session to delete.
        """
        ...

    def delete_checkpoint(self, session_name: str, timestamp: datetime) -> None:
        """Delete a specific checkpoint.

        Parameters
        ----------
        session_name : str
            The name of the session
        timestamp : datetime
            Timestamp of the checkpoint to delete

        Raises
        ------
        FileNotFoundError
            If checkpoint not found
        """
        ...

    def cleanup_old_checkpoints(
        self, session_name: str, keep_count: int = 5
    ) -> int:
        """Clean up old checkpoints, keeping only the most recent ones.

        Parameters
        ----------
        session_name : str
            Name of the session
        keep_count : int
            Number of recent checkpoints to keep
        """
        ...

    def clean_broken_symlinks(self, session_name: str | None = None) -> int:
        """Clean up broken symlinks.

        This cleans both internal symlinks (inside workspace) and
        registered external symlinks.

        Parameters
        ----------
        session_name : str | None
            If provided, clean only in that session's directory.
            If None, clean in all session directories.
        """
        ...

    def compact_registry(self) -> int:
        """Remove entries for non-existent checkpoints and links."""
        ...

    def verify_links(
        self, session_name: str | None = None
    ) -> Mapping[str, Iterable[str]]:
        """Verify all links are valid and pointing to correct targets.

        Parameters
        ----------
        session_name : str | None
            If provided, verify only links for that session.
            If None, verify all links.
        """
        ...

    def delete_checkpoints_batch(
        self, checkpoints: Iterable[tuple[str, datetime]]
    ) -> int:
        """Delete multiple checkpoints efficiently.

        Parameters
        ----------
        checkpoints : Iterable[tuple[str, datetime]]
            List of (session_name, timestamp) tuples
            Number of checkpoints successfully deleted
        """
        ...



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=line-too-long, broad-exception-caught,too-many-locals
# pylint: disable=too-many-try-statements,too-complex,too-many-branches
# pylint: disable=too-many-statements
# cspell: disable

"""Patch ag2 if needed."""

import argparse
import importlib.util
import re
import sys
from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from pathlib import Path


@dataclass
class Hunk:
    """Hunk class."""

    old_start: int
    old_len: int
    new_start: int
    new_len: int
    lines: list[str]


@dataclass
class FilePatch:
    """File patch class."""

    old_path: str | None
    new_path: str | None
    hunks: list[Hunk]


class PatchState(str, Enum):
    """Patch state."""

    CLEAN = "clean"
    ALREADY_APPLIED = "already_applied"
    DIVERGED = "diverged"


class PatchError(RuntimeError):
    """Patch error."""


_A_HDR = re.compile(r"^---\s+(?P<path>.+)")
_B_HDR = re.compile(r"^\+\+\+\s+(?P<path>.+)")
_HUNK = re.compile(r"^@@\s+-(\d+)(?:,(\d+))?\s+\+(\d+)(?:,(\d+))?\s+@@")


def _clean_header_path(raw: str) -> str:
    """Drop possible timestamps; normalize a/ and b/ prefixes."""
    # Header may be like: "path\tYYYY-MM-DD ...", or "a/foo/bar\t..."
    path = raw.strip()
    # Drop trailing metadata after a tab or space if it looks like timestamp
    if "\t" in path:
        path = path.split("\t", 1)[0]
    # Strip quotes if present
    if path.startswith('"') and path.endswith('"'):
        path = path[1:-1]
    # Normalize leading prefixes
    if path.startswith("a/"):
        path = path[2:]
    elif path.startswith("b/"):
        path = path[2:]
    return path


def _parse_unified_diff(text: str) -> list[FilePatch]:  # noqa: C901
    """Parse a unified diff (text-only) into FilePatch objects.

    Handles standard git-style '---'/'+++' headers and '@@' hunks.
    """
    lines = text.splitlines(keepends=True)
    i = 0
    patches: list[FilePatch] = []

    while i < len(lines):
        # Seek next file header
        if not lines[i].startswith("--- "):
            i += 1
            continue

        m_old = _A_HDR.match(lines[i].rstrip("\n"))
        if not m_old:
            raise PatchError(f"Malformed '---' header at line {i + 1}")
        i += 1
        if i >= len(lines) or not lines[i].startswith("+++ "):
            raise PatchError(f"Expected '+++' after '---' at line {i}")
        m_new = _B_HDR.match(lines[i].rstrip("\n"))
        if not m_new:
            raise PatchError(f"Malformed '+++' header at line {i + 1}")
        i += 1

        old_path_raw = m_old.group("path").strip()
        new_path_raw = m_new.group("path").strip()
        old_path = (
            None
            if old_path_raw == "/dev/null"
            else _clean_header_path(old_path_raw)
        )
        new_path = (
            None
            if new_path_raw == "/dev/null"
            else _clean_header_path(new_path_raw)
        )

        hunks: list[Hunk] = []
        # Collect hunks until next file or EOF
        while i < len(lines) and lines[i].startswith("@@ "):
            hm = _HUNK.match(lines[i].rstrip("\n"))
            if not hm:
                raise PatchError(
                    f"Malformed hunk header at line {i + 1}: {lines[i]}"
                )
            old_start = int(hm.group(1))
            old_len = int(hm.group(2) or "1")
            new_start = int(hm.group(3))
            new_len = int(hm.group(4) or "1")
            i += 1

            hunk_lines: list[str] = []
            # Hunk body: a sequence of lines starting with ' ', '+', '-', or '\
            # No newline...'
            while i < len(lines):
                line = lines[i]
                if (
                    line.startswith("--- ")
                    or line.startswith("@@ ")
                    or line.startswith("diff ")
                    or line.startswith("Index: ")
                ):
                    break
                if line.startswith("*** ") and " ****" in line:
                    break  # old-style context diff marker, treat as boundary
                # Unified diff content lines usually
                # start with ' ', '+', or '-'.
                # Special case: '\ No newline at end of file'
                if (
                    line.startswith(" ")
                    or line.startswith("+")
                    or line.startswith("-")
                    or line.startswith("\\")
                ):
                    hunk_lines.append(line)
                    i += 1
                else:
                    # End of hunk if unknown marker
                    break

            hunks.append(
                Hunk(old_start, old_len, new_start, new_len, hunk_lines)
            )

        patches.append(FilePatch(old_path, new_path, hunks))

    return patches


def _apply_hunks_to_content(content: list[str], hunks: list[Hunk]) -> list[str]:
    """Apply hunks to content (list of lines with keepends=True).

    Strict context matching (no fuzz).
    """
    # Work on a copy
    out = content[:]

    # Track offset between expected and actual positions after earlier edits
    line_offset = 0

    for h in hunks:
        # Position is 1-based in unified diff;
        # convert to 0-based index in our evolving buffer
        # Use the position for the *old* file to anchor context matching
        idx = h.old_start - 1 + line_offset

        # Build expected & replacement slices from the hunk
        expected_old: list[str] = []
        replacement_new: list[str] = []

        for line in h.lines:
            if line.startswith("\\"):
                # "\ No newline at end of file" — ignore during matching;
                #  we preserve original newline behavior
                continue
            tag = line[0]
            body = line[1:]
            if tag == " ":
                expected_old.append(body)
                replacement_new.append(body)
            elif tag == "-":
                expected_old.append(body)
            elif tag == "+":
                replacement_new.append(body)
            else:
                raise PatchError(f"Unknown hunk line tag: {tag!r}")

        # Verify the old slice matches exactly at idx
        old_slice = out[idx : idx + len(expected_old)]
        if old_slice != expected_old:
            # Provide a helpful preview
            preview_old = "".join(old_slice[:5])
            preview_exp = "".join(expected_old[:5])
            msg = (
                "Hunk context mismatch.\nExpected "
                f"at {idx}..{idx + len(expected_old)}:\n{preview_exp}\n"
                f"Found:\n{preview_old}\n"
                "The file likely diverged."
            )
            raise PatchError(msg)

        # Apply: replace that slice with the new lines
        out[idx : idx + len(expected_old)] = replacement_new

        # Update offset: new_len - old_len at this position
        line_offset += len(replacement_new) - len(expected_old)

    return out


def _find_package_root(package_name: str) -> Path:
    spec = importlib.util.find_spec(package_name)
    if spec is None:
        raise PatchError(
            f"Cannot find installed package/module '{package_name}'."
        )
    if spec.submodule_search_locations:
        pkg_dir = Path(next(iter(spec.submodule_search_locations)))
        return pkg_dir.parent
    if spec.origin:
        return Path(spec.origin).parent
    raise PatchError(
        f"Cannot resolve path for '{package_name}' (no origin or locations)."
    )


def _strip_components(path: str, strip: int) -> str:
    parts = Path(path).parts
    if strip <= 0:
        return str(Path(*parts))
    if strip >= len(parts):
        return ""
    return str(Path(*parts[strip:]))


def _guess_strip_for_package(paths: Iterable[str], package_name: str) -> int:
    for p in paths:
        parts = Path(p).parts
        if package_name in parts:
            return parts.index(package_name)
    return 0


def _apply_unified_diff_to_package(  # noqa: C901
    package_name: str,
    diff_path: Path | str,
    *,
    strip: int | None = None,
    dry_run: bool = True,
    allow_rejects: bool = True,
    encoding: str = "utf-8",
) -> None:
    """Apply a unified diff (text-only) to an installed package.

    - Supports creations (--- /dev/null), deletions (+++ /dev/null),
        and modifications.
    - Strict context matching; no fuzz.
    - Writes .rej files next to targets if allow_rejects=True on failure.
    - No external tools required.

    """
    diff_path = Path(diff_path).resolve()
    if not diff_path.exists():
        raise PatchError(f"Diff not found: {diff_path}")

    text = diff_path.read_text(encoding=encoding, errors="replace")
    file_patches: list[FilePatch] = _parse_unified_diff(text)

    # Determine strip if not provided
    if strip is None:
        candidate_paths: list[str] = []
        for fp in file_patches:
            if fp.new_path:
                candidate_paths.append(fp.new_path)
            elif fp.old_path:
                candidate_paths.append(fp.old_path)
        strip = _guess_strip_for_package(candidate_paths, package_name)

    pkg_root = _find_package_root(package_name)
    src_lines: list[str] = []

    # First pass: verify (dry run logic)
    for fp in file_patches:
        oldp = fp.old_path
        newp = fp.new_path

        # Target relative path after strip
        rel_old = _strip_components(oldp, strip) if oldp else None
        rel_new = _strip_components(newp, strip) if newp else None

        # Normalize paths for create/modify/delete
        is_create = (oldp is None) and (newp is not None)
        is_delete = (newp is None) and (oldp is not None)
        is_modify = (oldp is not None) and (newp is not None)

        target_rel = rel_new if not is_delete else rel_old
        if not target_rel:
            # Path stripped away; nothing to do
            # (but usually indicates wrong strip)
            msg = (
                f"After -p{strip}, target path is empty for file patch "
                f"(old={oldp!r}, new={newp!r}). Adjust strip level."
            )
            raise PatchError(msg)

        target_path = (pkg_root / target_rel).resolve()
        if is_create:
            # Validate create: file may or may not exist;
            # hunks should build new content
            # We simulate application starting from empty content
            src_lines = []
            new_lines = _apply_hunks_to_content(src_lines, fp.hunks)
            # No IO in dry-run
        elif is_delete:
            # Validate delete: file must exist and hunks must match content
            if not target_path.exists():
                raise PatchError(f"Delete target does not exist: {target_path}")
            src = target_path.read_text(encoding=encoding, errors="strict")
            src_lines = src.splitlines(keepends=True)
            _ = _apply_hunks_to_content(
                src_lines, fp.hunks
            )  # ensure it matches
        elif is_modify:
            # Modify
            if not target_path.exists():
                raise PatchError(f"Modify target does not exist: {target_path}")
            src = target_path.read_text(encoding=encoding, errors="strict")
            src_lines = src.splitlines(keepends=True)
            _ = _apply_hunks_to_content(src_lines, fp.hunks)

    # If only dry-run requested, stop here
    if dry_run:
        return

    # Second pass: actually mutate
    rejects: list[tuple[Path, str]] = []
    for fp in file_patches:
        oldp = fp.old_path
        newp = fp.new_path
        rel_old = _strip_components(oldp, strip) if oldp else None
        rel_new = _strip_components(newp, strip) if newp else None

        is_create = (oldp is None) and (newp is not None)
        is_delete = (newp is None) and (oldp is not None)
        is_modify = (oldp is not None) and (newp is not None)

        target_rel = rel_new if not is_delete else rel_old
        if not target_rel:
            continue
        target_path = (pkg_root / target_rel).resolve()
        try:
            if is_create:
                src_lines = []
                new_lines = _apply_hunks_to_content(src_lines, fp.hunks)
                target_path.parent.mkdir(parents=True, exist_ok=True)
                _safe_write_text(target_path, "".join(new_lines), encoding)
            elif is_delete:
                src = target_path.read_text(encoding=encoding, errors="strict")
                src_lines = src.splitlines(keepends=True)
                _ = _apply_hunks_to_content(src_lines, fp.hunks)
                # If it applies cleanly, delete
                target_path.unlink(missing_ok=False)
            elif is_modify:
                src = target_path.read_text(encoding=encoding, errors="strict")
                src_lines = src.splitlines(keepends=True)
                new_lines = _apply_hunks_to_content(src_lines, fp.hunks)
                _safe_write_text(target_path, "".join(new_lines), encoding)
        except Exception as e:  # collect rejects if asked
            if allow_rejects:
                rej = _format_reject(fp)
                rej_path = target_path.with_suffix(target_path.suffix + ".rej")
                rej_path.write_text(rej, encoding=encoding)
                rejects.append((rej_path, str(e)))
            else:
                raise

    if rejects:
        details = "\n".join(f"- {p}: {err}" for p, err in rejects)
        raise PatchError(
            "Some hunks failed to apply; .rej files were written:\n" + details
        )


def _safe_write_text(path: Path, data: str, encoding: str) -> None:
    # Write atomically
    tmp = path.with_suffix(path.suffix + ".tmp___apply")
    tmp.write_text(data, encoding=encoding)
    tmp.replace(path)


def _format_reject(fp: FilePatch) -> str:
    lines: list[str] = []
    op_old = fp.old_path or "/dev/null"
    op_new = fp.new_path or "/dev/null"
    lines.append(f"--- {op_old}\n")
    lines.append(f"+++ {op_new}\n")
    for h in fp.hunks:
        header = (
            f"@@ -{h.old_start},{h.old_len} +{h.new_start},{h.new_len} @@\n"
        )
        lines.append(header)
        lines.extend(h.lines)
    return "".join(lines)


def _dry_run_apply_parsed(
    package_name: str,
    file_patches: list[FilePatch],
    *,
    strip: int,
    encoding: str = "utf-8",
) -> bool:
    """Simulate a dry-run apply of already-parsed patches."""
    pkg_root = _find_package_root(package_name)
    try:
        for fp in file_patches:
            oldp = fp.old_path
            newp = fp.new_path
            rel_old = _strip_components(oldp, strip) if oldp else None
            rel_new = _strip_components(newp, strip) if newp else None

            is_create = (oldp is None) and (newp is not None)
            is_delete = (newp is None) and (oldp is not None)

            target_rel = rel_new if not is_delete else rel_old
            if not target_rel:
                return False
            target_path = (pkg_root / target_rel).resolve()
            if is_create:
                src_lines: list[str] = []
                _ = _apply_hunks_to_content(src_lines, fp.hunks)
            elif is_delete:
                if not target_path.exists():
                    return False
                src = target_path.read_text(encoding=encoding, errors="strict")
                src_lines = src.splitlines(keepends=True)
                _ = _apply_hunks_to_content(src_lines, fp.hunks)
            else:
                if not target_path.exists():
                    return False
                src = target_path.read_text(encoding=encoding, errors="strict")
                src_lines = src.splitlines(keepends=True)
                _ = _apply_hunks_to_content(src_lines, fp.hunks)
        return True
    except Exception:
        return False


def _reverse_file_patches(file_patches: list[FilePatch]) -> list[FilePatch]:
    """Produce a reversed patch.

    swap old/new paths, swap ranges, flip +/- lines.
    """
    rev: list[FilePatch] = []
    for fp in file_patches:
        rev_hunks: list[Hunk] = []
        for h in fp.hunks:
            flipped_lines: list[str] = []
            for ln in h.lines:
                if ln.startswith("\\"):
                    # keep "\ No newline..." lines as-is
                    flipped_lines.append(ln)
                elif ln.startswith("+"):
                    flipped_lines.append("-" + ln[1:])
                elif ln.startswith("-"):
                    flipped_lines.append("+" + ln[1:])
                else:
                    # context ' '
                    flipped_lines.append(ln)
            rev_hunks.append(
                Hunk(
                    old_start=h.new_start,
                    old_len=h.new_len,
                    new_start=h.old_start,
                    new_len=h.old_len,
                    lines=flipped_lines,
                )
            )
        rev.append(
            FilePatch(
                old_path=fp.new_path,
                new_path=fp.old_path,
                hunks=rev_hunks,
            )
        )
    return rev


def _check_patch_state(
    package_name: str,
    diff_path: Path | str,
    *,
    strip: int | None = None,
    encoding: str = "utf-8",
) -> PatchState:
    """Check whether a diff is clean to apply, already applied, or diverged."""
    diff_path = Path(diff_path).resolve()
    text = diff_path.read_text(encoding=encoding, errors="replace")
    parsed = _parse_unified_diff(text)

    if strip is None:
        candidate_paths: list[str] = []
        for fp in parsed:
            if fp.new_path:
                candidate_paths.append(fp.new_path)
            elif fp.old_path:
                candidate_paths.append(fp.old_path)
        strip = _guess_strip_for_package(candidate_paths, package_name)

    # Forward
    if _dry_run_apply_parsed(
        package_name, parsed, strip=strip, encoding=encoding
    ):
        return PatchState.CLEAN

    # Reverse (means target already matches post-image)
    reversed_parsed = _reverse_file_patches(parsed)
    if _dry_run_apply_parsed(
        package_name, reversed_parsed, strip=strip, encoding=encoding
    ):
        return PatchState.ALREADY_APPLIED

    return PatchState.DIVERGED


def apply_patch(
    package_name: str,
    diff_path: Path | str,
    *,
    strip: int | None = None,
    allow_rejects: bool = True,
    encoding: str = "utf-8",
    dry_run: bool = False,
) -> None:
    """Apply patch.

    Parameters
    ----------
    package_name : str
        The name of the package to patch.
    diff_path : Path | str
        The path of the .dff file with the changes.
    strip : int | None
        the -p level; if None we try to guess using the package name.
    allow_rejects : bool
        on real apply, write .rej files on hunk mismatch (still raises)
    encoding : str
        The text encoding for reading/writing files.
    dry_run : bool
        if True, validate first; if it passes, function returns (no mutation).

    1) Dry-run to validate.
    2) Real apply with .rej generation on any failure (and raise).
    """
    state = _check_patch_state(
        package_name, diff_path, strip=strip, encoding=encoding
    )
    if state == PatchState.CLEAN:
        _apply_unified_diff_to_package(
            package_name,
            diff_path,
            strip=strip,
            dry_run=True,
            allow_rejects=allow_rejects,
            encoding=encoding,
        )
        if not dry_run:
            _apply_unified_diff_to_package(
                package_name,
                diff_path,
                strip=strip,
                dry_run=False,
                allow_rejects=allow_rejects,
                encoding=encoding,
            )


def patch_ag2() -> None:
    """Patch ag2 if a diff file is found."""
    diff_path = Path(__file__).parent / "ag2.diff"
    if diff_path.is_file():
        try:
            apply_patch("autogen", diff_path)
        except BaseException as e:  # pylint: disable=broad-exception-caught
            print(e)


def _cli() -> argparse.ArgumentParser:
    """Get cli args parser."""
    parser = argparse.ArgumentParser(
        description=(
            "Apply a unified diff (text-only) to an installed Python package."
        )
    )

    parser.add_argument(
        "package",
        help="Import name of the installed package to patch (e.g. 'ag2').",
    )
    parser.add_argument(
        "diff", help="Path to unified diff file (produced without --binary)."
    )
    parser.add_argument(
        "-p",
        "--strip",
        type=int,
        default=None,
        help=(
            "Strip leading path components (like git apply -pN). "
            "If omitted, auto-guess using the package name."
        ),
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Only validate the diff; do not modify files.",
    )
    parser.add_argument(
        "--no-rejects",
        action="store_true",
        help=(
            "Fail immediately if a hunk cannot be applied; "
            "do not write .rej files."
        ),
    )
    parser.add_argument(
        "--encoding",
        default="utf-8",
        help="Encoding used to read/write source files. Default: utf-8",
    )
    parser.add_argument(
        "--state",
        action="store_true",
        help=(
            "Print whether the patch is clean, already_applied, or diverged."
        ),
    )

    return parser


def main() -> None:
    """Entry point for cli usage."""
    parser = _cli()
    args, _ = parser.parse_known_args()
    if args.state:
        st = _check_patch_state(
            args.package, args.diff, strip=args.strip, encoding=args.encoding
        )
        value = " ".join(st.value.split("_"))
        print(f"State: {value}")
        sys.exit(0)
    try:
        # Delegate to main engine
        apply_patch(
            package_name=args.package,
            diff_path=args.diff,
            strip=args.strip,
            dry_run=args.dry_run,
            allow_rejects=not args.no_rejects,
            encoding=args.encoding,
        )
    except PatchError as e:
        print("\nERROR applying diff:\n", file=sys.stderr)
        print(str(e), file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print("\nUNEXPECTED ERROR:\n", file=sys.stderr)
        print(repr(e), file=sys.stderr)
        sys.exit(2)
    else:
        if args.dry_run:
            print("Dry-run successful: patch is clean.")
        else:
            print("Patch applied successfully.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnnecessaryIsInstance=false,reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false,reportArgumentType=false
# pyright: reportReturnType=false

"""User input data models and validation."""

import json
import os
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from pydantic import BaseModel, Field, field_validator
from typing_extensions import Annotated

from ..utils import detect_media_type
from .constants import CONTENT_MAPPING, ContentMappingEntry, MediaContent
from .content.audio import AudioMediaContent
from .content.base import (
    AudioContent,
    FileContent,
    ImageContent,
    VideoContent,
)
from .content.file import FileMediaContent
from .content.image import ImageMediaContent, ImageUrlMediaContent
from .content.text import TextMediaContent
from .content.video import VideoMediaContent


class UserInputData(BaseModel):
    """User's input data model."""

    content: Annotated[
        MediaContent | list[MediaContent],
        Field(
            description="The content of the input data.",
            title="Content",
        ),
    ]

    def __str__(self) -> str:
        """Get the string representation of the content."""
        return self.to_string()

    def __repr__(self) -> str:
        """Get the string representation of the UserInputData."""
        return f"UserInputData(content={self.to_string()})"

    def to_string(
        self,
        uploads_root: Path | None = None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        if isinstance(self.content, list):
            return " ".join(
                [
                    item.to_string(uploads_root, base_name)
                    for item in self.content
                ]
            )
        return self.content.to_string(uploads_root, base_name)

    @classmethod
    def content_from_string(cls, value: str) -> MediaContent:
        """Convert a string to the appropriate MediaContent type.

        Parameters
        ----------
        value : str
            The input string

        Returns
        -------
        MediaContent
            the appropriate MediaContent.
        """
        try:
            parsed = json.loads(value)
        except json.JSONDecodeError:
            return TextMediaContent(type="text", text=value)
        if isinstance(parsed, str):
            try:
                parsed = json.loads(parsed)
            except json.JSONDecodeError:
                return TextMediaContent(type="text", text=parsed)
            if isinstance(parsed, str):  # pragma: no cover
                return TextMediaContent(type="text", text=parsed)
        return cls.validate_content(parsed)

    @classmethod
    def content_from_dict(cls, value: dict[str, Any]) -> MediaContent:
        """Convert a dictionary to the appropriate MediaContent type.

        Parameters
        ----------
        value : dict[str, Any]
            The input dictionary

        Returns
        -------
        MediaContent
            the appropriate MediaContent.

        Raises
        ------
        ValueError
            If the content type is not supported or
            if any required field is missing.
        """
        content_type = detect_media_type(value)

        # Get the mapping for the detected content type
        if content_type not in CONTENT_MAPPING:
            raise ValueError(f"Unsupported content type: {content_type}")

        # Get the mapping for the content type
        mapping = CONTENT_MAPPING[content_type]
        return cls._build_media_content(value, mapping, content_type)

    @classmethod
    def _build_media_content(
        cls,
        value: dict[str, Any],
        mapping: ContentMappingEntry,
        content_type: str,
    ) -> MediaContent:
        """Try to construct the appropriate MediaContent from the mapping."""
        for field in mapping["fields"]:
            if field not in value:
                continue

            raw_val = value[field]

            if field == mapping["required_field"]:
                # if we have direct mapping to the required field,
                # let's try to instantiate the class directly
                try:
                    return mapping["cls"](**{field: raw_val})
                except ValueError:
                    pass  # let's try to convert it

            converted = cls._convert_simple_content(
                raw_val, mapping["required_field"]
            )
            if converted is not None:  # pragma: no branch
                return mapping["cls"](
                    **{mapping["required_field"]: converted}  # type: ignore
                )

        msg = (
            "Missing required field for content type "
            f"'{content_type}' in: {value}"
        )
        raise ValueError(msg)

    @staticmethod
    def _convert_simple_content(
        raw_val: Any, target_field: str
    ) -> ImageContent | VideoContent | AudioContent | FileContent | None:
        """Convert a simple string to the appropriate content, if applicable."""
        if not isinstance(raw_val, str):  # pragma: no cover
            return None

        if target_field in ("image_url", "image"):
            return ImageContent(url=raw_val)
        if target_field == "video":
            return VideoContent(url=raw_val)
        if target_field == "audio":
            return AudioContent(url=raw_val)
        if target_field == "file":
            filename = extract_filename_from_path(raw_val)
            return FileContent(name=filename, url=raw_val)

        return None  # pragma: no cover

    # noinspection PyNestedDecorators
    @field_validator("content", mode="before")
    @classmethod
    def validate_content(cls, v: Any) -> MediaContent:  # noqa: C901,D102
        """Validate the input data content.

        Parameters
        ----------
        v: Any
            The input data content

        Returns
        -------
        MediaContent
            The validated content

        Raises
        ------
        ValueError
            If the content is not valid.
        """
        if isinstance(
            v,
            (
                TextMediaContent,
                ImageMediaContent,
                ImageUrlMediaContent,
                VideoMediaContent,
                AudioMediaContent,
                FileMediaContent,
            ),
        ):
            return v

        # If it's a string, check if it is a dumped one
        if isinstance(v, str):
            return cls.content_from_string(v)

        # If it's a dictionary, check if it has a 'type' field
        if isinstance(v, dict):
            return cls.content_from_dict(v)

        # If it's a list
        if isinstance(v, list):
            return [cls.validate_content(item) for item in v]  # type: ignore

        # Default fallback
        return TextMediaContent(type="text", text=str(v))


def extract_filename_from_path(path_or_url: str) -> str:
    """Extract the filename from a given path or URL.

    Parameters
    ----------
    path_or_url: str
        The path or URL from which to extract the filename.

    Returns
    -------
    str
        The extracted filename.
    """
    if "://" in path_or_url:
        # It's a URL - parse it properly
        parsed = urlparse(path_or_url)
        # Extract filename from the path component
        filename = os.path.basename(parsed.path)

        # Handle edge cases where path might be empty or end with /
        if not filename and parsed.path.endswith("/"):  # pragma: no cover
            # Try to get the last directory name
            path_parts = [p for p in parsed.path.split("/") if p]
            filename = path_parts[-1] if path_parts else parsed.netloc
        elif not filename:  # pragma: no cover
            # Fallback to using netloc (domain) if no path
            filename = parsed.netloc

        return filename
    # else:
    # Local file path
    return os.path.basename(path_or_url)  # pragma: no cover



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownMemberType=false, reportAttributeAccessIssue=false
# pyright: reportUnknownArgumentType=false, reportUnusedParameter=false
"""Base runner for Waldiez workflows."""

import importlib.util
import json
import shutil
import sys
import tempfile
import threading
import traceback as tb
from pathlib import Path
from types import ModuleType, TracebackType
from typing import Any

import aiofiles
from aiofiles.os import wrap
from anyio.from_thread import start_blocking_portal
from typing_extensions import Self, override

from waldiez.exporter import WaldiezExporter
from waldiez.logger import WaldiezLogger, get_logger
from waldiez.models import Waldiez
from waldiez.storage import StorageManager, WaldiezCheckpoint, safe_name

from .dir_utils import a_chdir, chdir
from .environment import reset_env_vars, set_env_vars
from .events_mixin import EventsMixin
from .exceptions import StopRunningException
from .protocol import WaldiezRunnerProtocol
from .requirements_mixin import RequirementsMixin
from .results_mixin import ResultsMixin


# pylint: disable=too-many-public-methods
# noinspection PyBroadException
class WaldiezBaseRunner(WaldiezRunnerProtocol, RequirementsMixin, ResultsMixin):
    """Base runner for Waldiez."""

    _structured_io: bool
    _output_path: str | Path | None
    _uploads_root: str | Path | None
    _dot_env_path: Path | None
    _running: bool
    _waldiez_file: Path
    _flow_name: str
    _storage_manager: StorageManager
    _checkpoint: WaldiezCheckpoint | None
    _output_dir: Path
    _logger: WaldiezLogger

    def __init__(
        self,
        waldiez: Waldiez,
        output_path: str | Path | None,
        uploads_root: str | Path | None,
        structured_io: bool,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Waldiez manager."""
        workspace_arg: str | None = kwargs.pop("workspace", None)
        if not isinstance(workspace_arg, str):
            workspace_arg = None
        WaldiezBaseRunner._running = False
        WaldiezBaseRunner._structured_io = structured_io
        WaldiezBaseRunner._output_path = output_path
        WaldiezBaseRunner._uploads_root = uploads_root
        WaldiezBaseRunner._dot_env_path = Path(dot_env) if dot_env else None
        WaldiezBaseRunner._flow_name = safe_name(waldiez.name)
        WaldiezBaseRunner._storage_manager = StorageManager(None, workspace_arg)
        WaldiezBaseRunner._waldiez = waldiez
        EventsMixin.set_input_function(input)
        EventsMixin.set_print_function(print)
        EventsMixin.set_send_function(print)
        EventsMixin.set_async(waldiez.is_async)
        RequirementsMixin.__init__(self)
        self._called_install_requirements = False
        self._exporter = WaldiezExporter(waldiez)
        self._stop_requested = threading.Event()
        self._last_results: list[dict[str, Any]] = []
        self._last_exception: Exception | None = None
        self._running_lock = threading.Lock()
        self._loaded_module: ModuleType | None = None
        logger = kwargs.get("logger")
        if isinstance(logger, WaldiezLogger):
            self._logger = logger
        else:
            self._logger = get_logger()
        WaldiezBaseRunner._logger = self._logger
        waldiez_file = kwargs.get("waldiez_file", "")
        if isinstance(waldiez_file, str) and waldiez_file:
            waldiez_file_path = Path(waldiez_file).resolve()
        elif isinstance(waldiez_file, Path):
            waldiez_file_path = waldiez_file.resolve()
        else:
            waldiez_file_path = waldiez.dump(to=output_path)
        if not waldiez_file_path or not waldiez_file_path.is_file():
            raise ValueError("Could not resolve a waldiez file path")
        WaldiezBaseRunner._waldiez_file = waldiez_file_path
        if (waldiez_file_path.parent / ".env").exists():
            WaldiezBaseRunner._dot_env_path = waldiez_file_path.parent / ".env"
        WaldiezBaseRunner._checkpoint = None
        checkpoint_arg = kwargs.get("checkpoint", "")
        if checkpoint_arg and isinstance(checkpoint_arg, str):
            WaldiezBaseRunner._checkpoint = WaldiezBaseRunner._init_checkpoint(
                checkpoint_arg=checkpoint_arg,
            )
        self._output_dir = WaldiezBaseRunner._init_output_dir(output_path)
        WaldiezBaseRunner._check_dot_env(self._output_dir)

    @staticmethod
    def _init_output_dir(output_path: str | Path | None) -> Path:
        if output_path:
            if str(output_path).endswith((".py", ".ipynb", ".waldiez")):
                return Path(output_path).parent
            return (
                Path(output_path)
                if Path(output_path).is_dir()
                else Path(output_path).parent
            )
        return Path.cwd()  # should change on ".run(...)"

    @staticmethod
    def _store_run_paths(tmp_dir: Path, output_file: Path) -> None:
        """Store the path of the module that is to be run."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            dir_path = StorageManager.default_root()
            dir_path.mkdir(parents=True, exist_ok=True)
            msg = json.dumps(
                {
                    "src": str(tmp_dir),
                    "dst": str(output_file),
                    "name": WaldiezBaseRunner._waldiez.name,
                }
            )
            with open(
                dir_path / ResultsMixin.RUN_DETAILS, "w", encoding="utf-8"
            ) as f:
                f.write(msg)
        except BaseException:
            WaldiezBaseRunner._logger.error(tb.format_exc())

    @staticmethod
    def _remove_run_paths() -> None:
        """Remove run paths."""
        ResultsMixin._cleanup()

    @staticmethod
    async def _a_store_run_paths(tmp_dir: Path, output_file: Path) -> None:
        """Store the path of the module that is to be run."""
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            dir_path = StorageManager.default_root()
            dir_path.mkdir(parents=True, exist_ok=True)
            msg = json.dumps(
                {
                    "src": str(tmp_dir),
                    "dst": str(output_file),
                    "name": WaldiezBaseRunner._waldiez.name,
                }
            )
            async with aiofiles.open(
                dir_path / ResultsMixin.RUN_DETAILS, "w", encoding="utf-8"
            ) as f:
                await f.write(msg)
        except BaseException:
            WaldiezBaseRunner._logger.error(tb.format_exc())

    @staticmethod
    def _check_dot_env(output_dir: Path) -> None:
        if (
            WaldiezBaseRunner._dot_env_path
            and WaldiezBaseRunner._dot_env_path.is_file()
        ):
            return
        if (output_dir / ".env").is_file():
            WaldiezBaseRunner._dot_env_path = output_dir / ".env"

    @staticmethod
    def _init_checkpoint(
        checkpoint_arg: str,
    ) -> WaldiezCheckpoint | None:
        checkpoint: WaldiezCheckpoint | None = None
        # pylint: disable=broad-exception-caught,too-many-try-statements
        session_name = WaldiezBaseRunner._flow_name
        checkpoint_id, history_index = StorageManager.parse_checkpoint_arg(
            checkpoint_arg
        )
        if checkpoint_id == "latest":
            try:
                info = WaldiezBaseRunner._storage_manager.get_latest_checkpoint(
                    session_name
                )
                if info:
                    checkpoint = WaldiezBaseRunner._storage_manager.load(
                        info, history_index=history_index
                    )
            except BaseException:
                pass
        else:
            try:
                checkpoint_ts = WaldiezCheckpoint.parse_timestamp(checkpoint_id)
                info = WaldiezBaseRunner._storage_manager.get(
                    session_name, checkpoint_ts
                )
                if info:
                    checkpoint = WaldiezBaseRunner._storage_manager.load(
                        info, history_index=history_index
                    )
            except BaseException:
                pass
        return checkpoint

    @override
    @staticmethod
    def print(*args: Any, **kwargs: Any) -> None:
        """Print a message to the output stream.

        Parameters
        ----------
        *args : Any
            Positional arguments to print.
        **kwargs : Any
            Keyword arguments to print.
        """
        if len(args) == 1 and isinstance(args[0], dict):
            arg = json.dumps(args[0], default=str, ensure_ascii=False)
            EventsMixin.do_print(arg, **kwargs)
        else:
            EventsMixin.do_print(*args, **kwargs)

    @override
    def is_running(self) -> bool:
        """Check if the workflow is currently running.

        Returns
        -------
        bool
            True if the workflow is running, False otherwise.
        """
        with self._running_lock:
            return WaldiezBaseRunner._running

    def _load_module(self, output_file: Path, temp_dir: Path) -> ModuleType:
        """Load the module from the waldiez file."""
        file_name = output_file.name
        module_name = file_name.replace(".py", "")
        spec = importlib.util.spec_from_file_location(
            module_name, temp_dir / file_name
        )
        if not spec or not spec.loader:
            raise ImportError("Could not import the flow")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if not hasattr(module, "main"):
            raise ImportError(
                "The waldiez file does not contain a main(...) function"
            )
        self._loaded_module = module
        return module

    def _before_run(
        self,
        output_file: Path,
        uploads_root: Path | None,
    ) -> Path:
        """Run before the flow execution."""
        self.log.info("Preparing workflow file: %s", output_file)
        temp_dir = Path(tempfile.mkdtemp(prefix="wlz-"))
        self._output_dir = temp_dir
        file_name = output_file.name
        with chdir(to=temp_dir):
            self._exporter.export(
                path=file_name,
                force=True,
                uploads_root=uploads_root,
                structured_io=WaldiezBaseRunner._structured_io,
            )
            if self.dot_env_path and self.dot_env_path.is_file():
                shutil.copyfile(
                    str(self.dot_env_path),
                    str(temp_dir / ".env"),
                )
        return temp_dir

    async def _a_before_run(
        self,
        output_file: Path,
        uploads_root: Path | None,
    ) -> Path:
        """Run before the flow execution asynchronously."""
        temp_dir = Path(tempfile.mkdtemp(prefix="wlz-"))
        self._output_dir = temp_dir
        file_name = output_file.name
        async with a_chdir(to=temp_dir):
            self._exporter.export(
                path=file_name,
                uploads_root=uploads_root,
                structured_io=self.structured_io,
                force=True,
            )
            if self.dot_env_path and self.dot_env_path.is_file():
                wrapped = wrap(shutil.copyfile)
                await wrapped(
                    str(self.dot_env_path),
                    str(temp_dir / ".env"),
                )
        return temp_dir

    def _run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow."""
        raise NotImplementedError(
            "The _run method must be implemented in the subclass."
        )

    async def _a_run(
        self,
        temp_dir: Path,
        output_file: Path,
        uploads_root: Path | None,
        skip_mmd: bool,
        skip_timeline: bool,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow asynchronously."""
        raise NotImplementedError(
            "The _a_run method must be implemented in the subclass."
        )

    def _after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        waldiez_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> Path | None:
        """Run after the flow execution."""
        # Save results
        self._last_results = results

        # Reset stop flag for next run
        self._stop_requested.clear()
        # pylint: disable=broad-exception-caught
        try:
            return ResultsMixin.post_run(
                results=results,
                error=error,
                temp_dir=temp_dir,
                output_file=output_file,
                flow_name=WaldiezBaseRunner._flow_name,
                waldiez_file=waldiez_file,
                uploads_root=uploads_root,
                skip_mmd=skip_mmd,
                skip_timeline=skip_timeline,
                storage_manager=WaldiezBaseRunner._storage_manager,
            )
        except BaseException as exc:  # pragma: no cover
            self.log.warning("Error occurred during after_run: %s", exc)
        self.log.info("Cleanup completed")
        return None

    async def _a_after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        waldiez_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> Path | None:
        """Run after the flow execution asynchronously."""
        self._last_results = results
        self._stop_requested.clear()
        # pylint: disable=broad-exception-caught
        try:
            return await ResultsMixin.a_post_run(
                results=results,
                error=error,
                temp_dir=temp_dir,
                output_file=output_file,
                flow_name=WaldiezBaseRunner._flow_name,
                waldiez_file=waldiez_file,
                uploads_root=uploads_root,
                skip_mmd=skip_mmd,
                skip_timeline=skip_timeline,
            )
        except BaseException as exc:  # pragma: no cover
            self.log.warning("Error occurred during a_after_run: %s", exc)
        self.log.info("Cleanup completed")
        return None

    @staticmethod
    def _prepare_paths(
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
    ) -> tuple[Path, Path | None]:
        """Prepare the output and uploads paths."""
        uploads_root_path: Path | None = None
        if uploads_root is not None:
            uploads_root_path = Path(uploads_root)
            WaldiezBaseRunner._uploads_root = uploads_root_path

        if output_path is not None:
            output_path = Path(output_path)
            WaldiezBaseRunner._output_path = output_path
        if not WaldiezBaseRunner._output_path:
            WaldiezBaseRunner._output_path = (
                Path.cwd() / safe_name(WaldiezBaseRunner._waldiez.name)
            ).with_suffix(".py")
        output_file: Path = Path(WaldiezBaseRunner._output_path)
        WaldiezBaseRunner._check_dot_env(output_file.parent)
        return output_file, uploads_root_path

    @override
    def before_run(
        self,
        output_file: Path,
        uploads_root: Path | None,
    ) -> Path:
        """Run the before_run method synchronously.

        Parameters
        ----------
        output_file : Path
            The path to the output file.
        uploads_root : Path | None
            The root path for uploads, if any.

        Returns
        -------
        Path
            The path to the temporary directory created before running the flow.
        """
        return self._before_run(
            output_file=output_file,
            uploads_root=uploads_root,
        )

    @override
    async def a_before_run(
        self,
        output_file: Path,
        uploads_root: Path | None,
    ) -> Path:
        """Run the _a_before_run method asynchronously.

        Parameters
        ----------
        output_file : Path
            The path to the output file.
        uploads_root : Path | None
            The root path for uploads, if any.

        Returns
        -------
        Path
            The path to the temporary directory created before running the flow.
        """
        return await self._a_before_run(
            output_file=output_file,
            uploads_root=uploads_root,
        )

    def prepare(
        self,
        output_path: str | Path | None,
        uploads_root: str | Path | None,
    ) -> tuple[Path, Path, Path | None]:
        """Prepare the paths and environment for running the flow.

        Parameters
        ----------
        output_path : str | Path | None
            The output path for the flow, by default None.
        uploads_root : str | Path | None
            The root path for uploads, by default None.

        Returns
        -------
        tuple[Path, Path, Path | None]
            A tuple containing:
            - The path to the output file.
            - The path to the temporary directory created for the run.
            - The root path for uploads, if specified, otherwise None.
        """
        output_file, uploads_root_path = self._prepare_paths(
            output_path=output_path,
            uploads_root=uploads_root,
        )
        temp_dir = self.before_run(
            output_file=output_file,
            uploads_root=uploads_root_path,
        )
        self.install_requirements()
        return temp_dir, output_file, uploads_root_path

    # noinspection PyProtocol
    # pylint: disable=too-many-locals,unused-argument
    @override
    def run(
        self,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow.

        Parameters
        ----------
        output_path : str | Path | None
            The output path, by default None.
        uploads_root : str | Path | None
            The runtime uploads root, by default None.
        structured_io : bool
            Whether to use structured IO instead of the default 'input/print'.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool
            Whether to skip generating the timeline JSON.
        dot_env : str | Path | None
            The path to the .env file, if any.
        **kwargs : Any
            Additional keyword arguments for the run method.

        Returns
        -------
        list[dict[str, Any]]
            The results of the run.

        Raises
        ------
        RuntimeError
            If the runner is already running, the workflow is not async,
            or an error occurs during the run.
        StopRunningException
            If the run is stopped by the user.
        """
        if dot_env is not None:
            resolved = Path(dot_env).resolve()
            if resolved.is_file():
                WaldiezBaseRunner._dot_env_path = resolved
        if structured_io is not None:
            WaldiezBaseRunner._structured_io = structured_io
        if self.is_running():
            raise RuntimeError("Workflow already running")
        if self.is_async:
            with start_blocking_portal(backend="asyncio") as portal:
                return portal.call(
                    self.a_run,
                    output_path,
                    uploads_root,
                    structured_io,
                    skip_mmd,
                )
        temp_dir, output_file, uploads_root_path = self.prepare(
            output_path=output_path,
            uploads_root=uploads_root,
        )
        WaldiezBaseRunner._running = True
        results: list[dict[str, Any]] = []
        error: BaseException | None = None
        old_env_vars = set_env_vars(self._waldiez.get_flow_env_vars())
        output_dir = output_file.parent
        try:
            with chdir(to=temp_dir):
                sys.path.insert(0, str(temp_dir))
                results = self._run(
                    temp_dir=temp_dir,
                    output_file=output_file,
                    uploads_root=uploads_root_path,
                    skip_mmd=skip_mmd,
                    skip_timeline=skip_timeline,
                )
        except (SystemExit, StopRunningException, KeyboardInterrupt) as exc:
            error = exc
            self.log.warning("Execution stopped: %s", exc)
            raise StopRunningException(StopRunningException.reason) from exc
        except BaseException as exc:  # pylint: disable=broad-exception-caught
            self.log.error("Error occurred while running workflow: %s", exc)
            error = exc
        finally:
            WaldiezBaseRunner._running = False
            reset_env_vars(old_env_vars)
            output = self.after_run(
                results=results,
                error=error,
                output_file=output_file,
                uploads_root=uploads_root_path,
                temp_dir=temp_dir,
                skip_mmd=skip_mmd,
                skip_timeline=skip_timeline,
            )
            if output:
                output_dir = output
        EventsMixin.do_print("<Waldiez> - Done running the flow.")
        if sys.path[0] == str(temp_dir):
            sys.path.pop(0)
        return self.get_results(results, output_dir)

    async def a_prepare(
        self,
        output_path: str | Path | None,
        uploads_root: str | Path | None,
    ) -> tuple[Path, Path, Path | None]:
        """Prepare the paths for the async run.

        Parameters
        ----------
        output_path : str | Path | None
            The output path, by default None.
        uploads_root : str | Path | None
            The uploads root path, by default None.

        Returns
        -------
        tuple[Path, Path, Path | None]
            The temporary directory, output file, and uploads root path.
        """
        output_file, uploads_root_path = self._prepare_paths(
            output_path=output_path,
            uploads_root=uploads_root,
        )
        temp_dir = await self._a_before_run(
            output_file=output_file,
            uploads_root=uploads_root_path,
        )
        await self.a_install_requirements()
        return temp_dir, output_file, uploads_root_path

    # noinspection DuplicatedCode
    # noinspection PyProtocol
    @override
    async def a_run(
        self,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool | None = None,
        skip_mmd: bool = False,
        skip_timeline: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        """Run the Waldiez flow asynchronously.

        Parameters
        ----------
        output_path : str | Path | None
            The output path, by default None.
        uploads_root : str | Path | None
            The runtime uploads root.
        structured_io : bool
            Whether to use structured IO instead of the default 'input/print'.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool
            Whether to skip generating the timeline JSON.
        dot_env : str | Path | None
            The path to the .env file, if any.
        **kwargs : Any
            Additional keyword arguments for the a_run method.

        Returns
        -------
        list[dict[str, Any]]
            The results of the run.

        Raises
        ------
        RuntimeError
            If the runner is already running, the workflow is not async
            or an error occurs during the run.
        StopRunningException
            If the run is stopped by the user.
        """
        if dot_env is not None:
            resolved = Path(dot_env).resolve()
            if resolved.is_file():
                WaldiezBaseRunner._dot_env_path = resolved
        if structured_io is not None:
            WaldiezBaseRunner._structured_io = structured_io
        if self.is_running():
            raise RuntimeError("Workflow already running")
        temp_dir, output_file, uploads_root_path = await self.a_prepare(
            output_path=output_path,
            uploads_root=uploads_root,
        )
        WaldiezBaseRunner._running = True
        results: list[dict[str, Any]] = []
        error: BaseException | None = None
        output_dir = output_file.parent
        old_env_vars = set_env_vars(self._waldiez.get_flow_env_vars())
        try:
            async with a_chdir(to=temp_dir):
                sys.path.insert(0, str(temp_dir))
                results = await self._a_run(
                    temp_dir=temp_dir,
                    output_file=output_file,
                    uploads_root=uploads_root_path,
                    skip_mmd=skip_mmd,
                    skip_timeline=skip_timeline,
                )
        except (SystemExit, StopRunningException, KeyboardInterrupt) as exc:
            self.log.warning("Execution stopped: %s", exc)
            error = exc
            raise StopRunningException(StopRunningException.reason) from exc
        except BaseException as exc:  # pylint: disable=broad-exception-caught
            self.log.error("Error occurred while running workflow: %s", exc)
            error = exc
        finally:
            WaldiezBaseRunner._running = False
            reset_env_vars(old_env_vars)
            output = await self.a_after_run(
                results=results,
                error=error,
                output_file=output_file,
                uploads_root=uploads_root_path,
                temp_dir=temp_dir,
                skip_mmd=skip_mmd,
                skip_timeline=skip_timeline,
            )
            if output:
                output_dir = output
        if sys.path[0] == str(temp_dir):
            sys.path.pop(0)
        return await self.a_get_results(results, output_dir)

    @override
    def after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> Path | None:
        """Actions to perform after running the flow.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The results of the flow run.
        error : BaseException | None
            Optional error during the run.
        output_file : Path
            The path to the output file.
        uploads_root : Path | None
            The root path for uploads, if any.
        temp_dir : Path
            The path to the temporary directory used during the run.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool
            Whether to skip generating the timeline JSON.

        Returns
        -------
        Path | None
            The destination directory if output file, else None
        """
        return self._after_run(
            results=results,
            error=error,
            output_file=output_file,
            waldiez_file=WaldiezBaseRunner._waldiez_file,
            uploads_root=uploads_root,
            temp_dir=temp_dir,
            skip_mmd=skip_mmd,
            skip_timeline=skip_timeline,
        )

    @override
    async def a_after_run(
        self,
        results: list[dict[str, Any]],
        error: BaseException | None,
        output_file: Path,
        uploads_root: Path | None,
        temp_dir: Path,
        skip_mmd: bool,
        skip_timeline: bool,
    ) -> Path | None:
        """Asynchronously perform actions after running the flow.

        Parameters
        ----------
        results : list[dict[str, Any]]
            The results of the flow run.
        error : BaseException | None
            Optional error during the run.
        output_file : Path
            The path to the output file.
        uploads_root : Path | None
            The root path for uploads, if any.
        temp_dir : Path
            The path to the temporary directory used during the run.
        skip_mmd : bool
            Whether to skip generating the mermaid diagram.
        skip_timeline : bool

        Returns
        -------
        Path | None
            The destination directory if output file, else None
        """
        return await self._a_after_run(
            results=results,
            error=error,
            output_file=output_file,
            uploads_root=uploads_root,
            temp_dir=temp_dir,
            skip_mmd=skip_mmd,
            skip_timeline=skip_timeline,
            waldiez_file=WaldiezBaseRunner._waldiez_file,
        )

    @property
    def waldiez(self) -> Waldiez:
        """Get the Waldiez instance."""
        return WaldiezBaseRunner._waldiez

    @property
    def waldiez_file(self) -> Path:
        """Get the path to the waldiez file."""
        return self._waldiez_file

    @property
    def is_async(self) -> bool:
        """Check if the workflow is async."""
        return WaldiezBaseRunner._waldiez.is_async

    @property
    def running(self) -> bool:
        """Get the running status."""
        return self.is_running()

    @property
    def log(self) -> WaldiezLogger:
        """Get the logger for the runner."""
        return self._logger

    @property
    def structured_io(self) -> bool:
        """Check if the runner is using structured IO."""
        return WaldiezBaseRunner._structured_io

    @property
    def dot_env_path(self) -> Path | None:
        """Get the path to the .env file, if any."""
        return WaldiezBaseRunner._dot_env_path

    @property
    def output_path(self) -> str | Path | None:
        """Get the output path for the runner."""
        return WaldiezBaseRunner._output_path

    @property
    def uploads_root(self) -> str | Path | None:
        """Get the uploads root path for the runner."""
        return WaldiezBaseRunner._uploads_root

    @property
    def state_json(self) -> Path | None:
        """Get the state.json path to resume from if any."""
        if WaldiezBaseRunner._checkpoint:
            state_file = WaldiezBaseRunner._checkpoint.state_file
            if state_file.is_file():
                return state_file
        return None

    @classmethod
    def load(
        cls,
        waldiez_file: str | Path,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        requirements: list[str] | None = None,
        output_path: str | Path | None = None,
        uploads_root: str | Path | None = None,
        structured_io: bool = False,
        dot_env: str | Path | None = None,
        **kwargs: Any,
    ) -> "WaldiezBaseRunner":
        """Load a waldiez flow from a file and create a runner.

        Parameters
        ----------
        waldiez_file : str | Path
            The path to the waldiez file.
        name : str | None, optional
            The name of the flow, by default None.
        description : str | None, optional
            The description of the flow, by default None.
        tags : list[str] | None, optional
            The tags for the flow, by default None.
        requirements : list[str] | None, optional
            The requirements for the flow, by default None.
        output_path : str | Path | None, optional
            The path to save the output file, by default None.
        uploads_root : str | Path | None, optional
            The root path for uploads, by default None.
        structured_io : bool, optional
            Whether to use structured IO instead of the default 'input/print',
            by default False.
        dot_env : str | Path | None, optional
            The path to the .env file, if any, by default None.
        **kwargs : Any
            Additional kwargs to pass.

        Returns
        -------
        WaldiezBaseRunner
            An instance of WaldiezBaseRunner initialized with the loaded flow.
        """
        waldiez = Waldiez.load(
            waldiez_file,
            name=name,
            description=description,
            tags=tags,
            requirements=requirements,
        )
        return cls(
            waldiez=waldiez,
            output_path=output_path,
            uploads_root=uploads_root,
            structured_io=structured_io,
            dot_env=dot_env,
            **kwargs,
        )

    def stop(self) -> None:
        """Stop the workflow execution."""
        self._stop_requested.set()

    def is_stop_requested(self) -> bool:
        """Check if a stop has been requested.

        Returns
        -------
        bool
            True if stop has been requested, False otherwise.
        """
        return self._stop_requested.is_set()

    def set_stop_requested(self) -> None:
        """Set the stop requested flag."""
        self._stop_requested.set()

    def __enter__(self) -> Self:
        """Enter the context manager."""
        return self

    async def __aenter__(self) -> Self:
        """Enter the context manager asynchronously."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager."""
        if self.is_running():
            self._stop_requested.set()

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager asynchronously."""
        if self.is_running():
            self._stop_requested.set()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=line-too-long
# flake8: noqa: E501
"""Common utilities for exporting chats."""

import json

from waldiez.models import WaldiezAgent, WaldiezChat


def get_chat_message_string(
    sender: WaldiezAgent,
    chat: WaldiezChat,
    chat_names: dict[str, str],
) -> tuple[str, str | None]:
    """Get the agent's message as a string.

    Parameters
    ----------
    sender : WaldiezAgent
        The sender.
    chat : WaldiezChat
        The chat.
    chat_names : dict[str, str]
        A mapping of chat id to chat name with all the chats in the flow.

    Returns
    -------
    tuple[str, Optional[str]]
        If the message is a string, the message content and None.
        If the message is a method, the method name and the method content.
        If the message is None, 'None' and None.
    """
    if not chat.message or chat.message.type == "none":
        return "None", None
    if chat.message.type == "string":
        if chat.message.content is None:  # pragma: no cover
            # should be covered previously on pydantic validation
            return "None", None
        if not chat.message.content:
            return "", None
        return json.dumps(chat.message.content, ensure_ascii=False), None

    is_rag_with_carryover = sender.is_rag_user and chat.message.use_carryover
    chat_name = chat_names[chat.id]
    function_content, function_name = chat.get_message_function(
        name_suffix=chat_name,
        is_rag=is_rag_with_carryover,
    )
    return function_name, function_content


def _stop_logging(space: str, is_async: bool) -> str:
    tab = "    "
    if is_async:
        return f"{space}{tab}{tab}{tab}{tab}await stop_logging()\n"
    return f"{space}{tab}{tab}{tab}{tab}stop_logging()\n"


def get_event_handler_string(
    space: str,
    is_async: bool,
) -> str:
    """Get the event handler string.

    Parameters
    ----------
    space : str
        The space before the content.
    is_async : bool
        Whether the handler is asynchronous.

    Returns
    -------
    str
        The event handler string.
    """
    tab = "    "
    content = (
        f"{space}if not isinstance(results, list):\n"
        f"{space}{tab}results = [results]  # pylint: disable=redefined-variable-type\n"
        f"{space}got_agents = False\n"
        f"{space}known_agents: list[ConversableAgent] = []\n"
        f"{space}result_events: list[dict[str, Any]] = []\n"
        f"{space}if on_event:\n"
        f"{space}{tab}for index, result in enumerate(results):\n"
        f"{space}{tab}{tab}result_events = []\n"
    )
    await_str = "await " if is_async else ""
    if is_async:
        content += (
            f"{space}{tab}{tab}async for event in result.events:\n{_add_event_dump(space, tab)}"
            f"{space}{tab}{tab}{tab}if not got_agents:\n"
            f"{space}{tab}{tab}{tab}{tab}known_agents = _get_known_agents()\n"
            f"{space}{tab}{tab}{tab}{tab}got_agents = True\n"
            f"{space}{tab}{tab}{tab}a_pause_event.clear()\n"
            f"{space}{tab}{tab}{tab}try:\n"
            f"{space}{tab}{tab}{tab}{tab}should_continue = await on_event(event, known_agents)\n"
            f"{space}{tab}{tab}{tab}{tab}a_pause_event.set()\n"
            f"{space}{tab}{tab}{tab}except BaseException as e:\n{_stop_logging(space, is_async)}"
            f"{space}{tab}{tab}{tab}{tab}await store_error(e)\n"
            f"{space}{tab}{tab}{tab}{tab}raise SystemExit(\n"
            f'{space}{tab}{tab}{tab}{tab}{tab}"Error in event handler: " + str(e)\n'
            f"{space}{tab}{tab}{tab}{tab}) from e\n"
        )
    else:
        content += (
            f"{space}{tab}{tab}for event in result.events:\n{_add_event_dump(space, tab)}"
            f"{space}{tab}{tab}{tab}if not got_agents:\n"
            f"{space}{tab}{tab}{tab}{tab}known_agents = _get_known_agents()\n"
            f"{space}{tab}{tab}{tab}{tab}got_agents = True\n"
            f"{space}{tab}{tab}{tab}pause_event.clear()\n"
            f"{space}{tab}{tab}{tab}try:\n"
            f"{space}{tab}{tab}{tab}{tab}should_continue = on_event(event, known_agents)\n"
            f"{space}{tab}{tab}{tab}{tab}pause_event.set()\n"
            f"{space}{tab}{tab}{tab}except BaseException as e:\n{_stop_logging(space, is_async)}"
            f"{space}{tab}{tab}{tab}{tab}store_error(e)\n"
            f"{space}{tab}{tab}{tab}{tab}raise SystemExit(\n"
            f'{space}{tab}{tab}{tab}{tab}{tab}"Error in event handler: " + str(e)\n'
            f"{space}{tab}{tab}{tab}{tab}) from e\n"
        )
    content += (
        f'{space}{tab}{tab}{tab}if getattr(event, "type") == "run_completion":\n'
        f"{space}{tab}{tab}{tab}{tab}break\n"
        f"{space}{tab}{tab}{tab}if not should_continue:\n{_stop_logging(space, is_async)}"
        f"{space}{tab}{tab}{tab}{tab}{await_str}store_error()\n"
        f'{space}{tab}{tab}{tab}{tab}raise SystemExit("Event handler stopped processing")\n'
    )
    content += get_result_dicts_string(space, is_async)
    content += (
        f"{space}else:\n{space}{tab}for index, result in enumerate(results):\n"
        f"{space}{tab}{tab}result_events = []\n"
    )
    if is_async:
        content += (
            f"{space}{tab}{tab}await result.process()\n"
            f"{space}{tab}{tab}async for event in result.events:\n"
        )
    else:
        content += (
            f"{space}{tab}{tab}result.process()\n"
            f"{space}{tab}{tab}for event in result.events:\n"
        )
    content += _add_event_dump(space, tab)
    content += get_result_dicts_string(space, is_async)
    return content


def _add_event_dump(space: str, tab: str) -> str:
    return (
        f"{space}{tab}{tab}{tab}try:\n"
        f'{space}{tab}{tab}{tab}{tab}result_events.append(event.model_dump(mode="json", fallback=str))\n'
        f"{space}{tab}{tab}{tab}except BaseException:  # pylint: disable=broad-exception-caught\n"
        f"{space}{tab}{tab}{tab}{tab}pass\n"
    )


def get_result_dicts_string(tab: str, is_async: bool) -> str:
    """Get the result dicts string.

    Parameters
    ----------
    tab : str
        The space string to use for indentation.
    is_async : bool
        Whether the function is asynchronous.

    Returns
    -------
    str
        The result dicts string.
    """
    # fmt: off
    space = f"{tab}        "
    flow_content = f"{space}result_cost = "
    if is_async:
        flow_content += "await "
    flow_content += "result.cost\n"
    flow_content += f"{space}result_context_variables = "
    if is_async:
        flow_content += "await "
    flow_content += "result.context_variables\n"
    flow_content += f"{space}result_dict = {{\n"
    flow_content += f'{space}    "index": index,\n'
    flow_content += f'{space}    "uuid": str(result.uuid),\n'
    flow_content += f'{space}    "events": result_events,\n'
    if is_async:
        flow_content += f'{space}    "messages": await result.messages,\n'
        flow_content += f'{space}    "summary": await result.summary,\n'
        flow_content += f'{space}    "cost": result_cost.model_dump(mode="json", fallback=str) if result_cost else None,\n'
        flow_content += f'{space}    "context_variables": result_context_variables.model_dump(mode="json", fallback=str) if result_context_variables else None,\n'
        flow_content += f'{space}    "last_speaker": await result.last_speaker,\n'
    else:
        flow_content += f'{space}    "messages": result.messages,\n'
        flow_content += f'{space}    "summary": result.summary,\n'
        flow_content += f'{space}    "cost": result_cost.model_dump(mode="json", fallback=str) if result_cost else None,\n'
        flow_content += f'{space}    "context_variables": result_context_variables.model_dump(mode="json", fallback=str) if result_context_variables else None,\n'
        flow_content += f'{space}    "last_speaker": result.last_speaker,\n'
    flow_content += f"{space}}}\n"
    flow_content += f"{space}result_dicts.append(result_dict)\n"
    # fmt: on
    return flow_content
